import os
from datetime import date
from inspect import stack
from logging import getLogger, Logger
from pathlib import Path
from unittest import TestCase

import xlsxwriter
from jsonschema import ValidationError
from pandas import DataFrame

from com_enovation.toolbox.excel_dashboard.config_checker import ConfigChecker
from com_enovation.toolbox.excel_dashboard.widgets.table_zoom.widget_table_zoom import WidgetTableZoom


class TestWidget(TestCase):

    _logger: Logger = getLogger(__name__)

    """
    We want to print a table of this kind
    
           |---------------------------------------------------------------|
           |- A ---------|- B ---------|- C ---------|- D ---------|- E ---|         
    -----                 -------------------------------------------------
    | 1 |    project ABC |  1-Jan-2022 |  1-Aug-2022 | 28-Aug-2022 | Trend |
    -----  --------------|-------------|-------------|-------------|-------|
    | 2 |  | weight      |     100 pts |     110 pts |     110 pts | Up    |
    -----  |-------------|-----------------------------------------|-------|
    | 3 |  | done        |      50 pts |      75 pts |      70 pts | Up    |
    -----  |-------------|-----------------------------------------|       |
    | 4 |  | progress    |         50% |         68% |         70% |       |
    -----  |-------------|-----------------------------------------|-------|
    | 5 |  | velocity    |              xxx pts/ mth | xxx pts/mth | Down  |
    -----  |-------------|-----------------------------------------|-------|
    | 6 |  | ***         | ***         | ***         | ***         | ***   |
    -----   ---------------------------------------------------------------
    """

    _test_data_by_pattern: list = [

        # Default values
        {"msg": "Default 1: default value", "test": {"value": 1.1}},
        {"msg": "Default 2: default parameter", "test": {"parameter": "Default 2"}},
        {
            "msg": "Default 3: EXCEPTION: both default value and default parameter",
            "test": {"value": 1.3, "parameter": "Default 3"},
            "exception": "We cannot have both default 'value' and default 'parameter' at the same time"
        },
        {
            "msg": "Default 4: EXCEPTION: empty",
            "test": {},
            "exception": "does not have enough properties"
        },

        # Pattern 1: one single cell, no list
        {
            "msg": "Pattern 1a: one single cell, no list - value",
            "test": {"C5": {"value": "Pattern 1a"}}
        },
        {
            "msg": "Pattern 1b: one single cell, no list - parameter",
            "test": {"C5": {"parameter": "Pattern 1b"}}
        },
        {
            "msg": "Pattern 1c: one single cell, no list - EXCEPTION: unexpected tag",
            "test": {"C5": {"format": {"bg_color": "black"}, "value": 1.3, "jsg": "Pattern 1c"}},
            "exception": "Additional properties are not allowed"
        },
        {
            "msg": "Pattern 1d: one single cell, no list - EXCEPTION: both value and parameter",
            "test": {"C5": {"format": {"bg_color": "black"}, "value": "Pattern 1d", "parameter": "Pattern 1d"}},
            "exception": "should not be valid under"
        },
        {
            "msg": "Pattern 1e: one single cell, no list - complex one",
            "test": {
                "format": {"bg_color": "gray", "fg_color": "white"},
                "parameter": "Pattern 1e",
                "conditional_formattings": [
                    {
                        "range": "@A@1:@A2",
                        "options": {
                            "type": "cell", "criteria": "equal to", "value": '"Failed"',
                            "format": {'bold': True, 'font_color': 'red'}
                        }
                    }
                ],
                "C5": {"format": {"bg_color": "black"}, "value": "Pattern 1e"}
            }
        },
        {
            "msg": "Pattern 1f: one single cell, no list - EXCEPTION: merge not allowed",
            "test": {"C5": {"format": {"bg_color": "black"}, "value": "Pattern 1d", "merged": True}},
            "exception": "should not be valid under"
        },
        {
            "msg": "Pattern 1g: one single cell, no list - one format only",
            "test": {"C5": {"format": {"bg_color": "black"}}},
        },
        {
            "msg": "Pattern 1h: one single cell, no list - EXCEPTION: empty cell",
            "test": {"C5": {}},
            "exception": "does not have enough properties"
        },

        # Pattern 2: range which replicates one cell
        {
            "msg": "Pattern 2a: range which replicates one cell - value",
            "test": {"C5:D@6": {"format": {"bg_color": "black"}, "value": "Pattern 2a"}}
        },
        {
            "msg": "Pattern 2b: range which replicates one cell - parameter",
            "test": {"C5:D@6": {"format": {"bg_color": "black"}, "parameter": "Pattern 2b"}}
        },
        {
            "msg": "Pattern 2c: range which replicates one cell - value, merged",
            "test": {"C5:D@6": {"format": {"bg_color": "black"}, "value": "Pattern 2c", "merged": True}}
        },
        {
            "msg": "Pattern 2d: range which replicates one cell - parameter, merged",
            "test": {"C5:D@6": {"format": {"bg_color": "black"}, "parameter": "Pattern 2d", "merged": True}}
        },
        {
            "msg": "Pattern 2e: range which replicates one cell - EXCEPTION: unexpected tag",
            "test": {"C5:D@6": {"format": {"bg_color": "black"}, "jsg": "Pattern 2e", "merged": True}},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": "Pattern 2f: range which replicates one cell - EXCEPTION: wrong parameter format",
            "test": {"C5:D@6": {"format": {"bg_color": "black"}, "parameter": 3.5, "merged": True}},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": "Pattern 2g: range which replicates one cell - EXCEPTION: empty cell",
            "test": {"C5:D@6": {}},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": "Pattern 2h: range which replicates one cell - only merge",
            "test": {"C5:D@6": {"merged": True}}
        },

        # Pattern 3: range as list of cells
        {
            "msg": "Pattern 3a: range as list of cells",
            "test": {"C5:D@6": [
                    {"format": {"border": 8, "border_color": "red"}, "value": "Pattern 3a"},
                    {"format": {"border": 8, "border_color": "red"}, "parameter": "Pattern 3a"}
            ]}
        },
        {
            "msg": "Pattern 3b: range as list of cells",
            "test": {"C5:D@6": [{"value": "Pattern 3b"}, {"parameter": "Pattern 3b"}]}
        },
        {
            "msg": "Pattern 3c: range as list of cells - EXCEPTION: empty list",
            "test": {"C5:D@6": []},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": "Pattern 3d: range as list of cells - EXCEPTION: empty cell in the list",
            "test": {"C5:D@6": [{}]},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": "Pattern 3e: range as list of cells - EXCEPTION: non cell in the list",
            "test": {"C5:D@6": [3, 2, 1]},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": "Pattern 3e: range as list of cells - EXCEPTION: merge not expected",
            "test": {"C5:D@6": [{"value": "Pattern 3e", "merged": True}]},
            "exception": "is not valid under any of the given schemas"
        },

        # Pattern 4: range with one value or parameter and a list of formats
        {
            "msg": "Pattern 4a: range with one value or parameter and a list of formats - one value only",
            "test": {"C5:D@6": {"value": "pattern 4a"}}
        },
        {
            "msg": "Pattern 4b: range with one value or parameter and a list of formats - one parameter only",
            "test": {"C5:D@6": {"parameter": "pattern 4b"}}
        },
        {
            "msg": "Pattern 4c: range with one value or parameter and a list of formats - list of formats only",
            "test": {"C5:D@6": {"format": [{"border": 8, "border_color": "red"}]}}
        },
        {
            "msg": "Pattern 4d: range with one value or parameter and a list of formats - one value & list of formats",
            "test": {"C5:D@6": {"value": 4.4, "format": [{"border": 8, "border_color": "red"}]}}
        },
        {
            "msg": "Pattern 4e: range with one value or parameter and a list of formats - one param & list of formats",
            "test": {"C5:D@6": {"parameter": "pattern 4e", "format": [{"border": 8, "border_color": "red"}]}}
        },
        {
            "msg": "Pattern 4f: range with one value or parameter and a list of formats - "
                   "EXCEPTION: empty list of formats",
            "test": {"C5:D@6": {"parameter": "pattern 4f", "format": []}},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": "Pattern 4g: range with one value or parameter and a list of formats - "
                   "EXCEPTION: list of formats containing a non format object",
            "test": {"C5:D@6": {"parameter": "pattern 4g", "format": [1, 2, 3]}},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": f"Pattern 4: range with format and value as: "
                   f"(i) Either one value to replicate (similar to pattern 2)"
                   f"(ii) Or list of values to apply (similar to patter 3)",
            "test": {
                "@B@1:@D@1": {
                    "format": {"border": 8, "border_color": "red"},
                    "value": ["date_1", "date_2", "date_3"]
                },
            }
        },

        # Pattern 5: range with list of values or parameters and a list of formats
        {
            "msg": "Pattern 5a: range with one value or parameter and a list of formats - list of values only",
            "test": {"C5:D@6": {"value": [5.1]}}
        },
        {
            "msg": "Pattern 5b: range with one value or parameter and a list of formats - list of parameters only",
            "test": {"C5:D@6": {"parameter": ["pattern 5b"]}}
        },
        {
            "msg": "Pattern 5c: range with one value or parameter and a list of formats - "
                   "list of values and list of formats",
            "test": {"C5:D@6": {"value": [5.3], "format":[{"border": 8, "border_color": "red"}]}}
        },
        {
            "msg": "Pattern 5d: range with one value or parameter and a list of formats - "
                   "list of parameters and list of formats",
            "test": {"C5:D@6": {"parameter": ["pattern 5d"], "format":[{"border": 8, "border_color": "red"}]}}
        },
        {
            "msg": "Pattern 5e: range with one value or parameter and a list of formats - "
                   "EXCEPTION: empty list of values",
            "test": {"C5:D@6": {"value": []}},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": "Pattern 5f: range with one value or parameter and a list of formats - "
                   "EXCEPTION: empty list of parameters",
            "test": {"C5:D@6": {"parameter": []}},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": "Pattern 5g: range with one value or parameter and a list of formats - "
                   "EXCEPTION: empty list of formats",
            "test": {"C5:D@6": {"format": []}},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": "Pattern 5h: range with one value or parameter and a list of formats - "
                   "EXCEPTION: weird value",
            "test": {"C5:D@6": {"value": [{"jsg": 3}]}},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": "Pattern 5i: range with one value or parameter and a list of formats - "
                   "EXCEPTION: weird parameter",
            "test": {"C5:D@6": {"parameter": [3.4]}},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": "Pattern 5j: range with one value or parameter and a list of formats - "
                   "EXCEPTION: weird format",
            "test": {"C5:D@6": {"format": [{"jsg": 3}]}},
            "exception": "is not valid under any of the given schemas"
        },

        # Pattern 6: range with list of values or list of parameters and one format
        {
            "msg": "Pattern 6a: range with list of values or list of parameters and one format -"
                   "List of values and one format",
            "test": {"C5:D@6": {"value": ["pattern 6a", 2, 3], "format": {"border": 8, "border_color": "red"}}}
        },
        {
            "msg": "Pattern 6b: range with list of values or list of parameters and one format -"
                   "List of parameters and one format",
            "test": {"C5:D@6": {"parameter": ["pattern", "6a", "3"], "format": {"border": 8, "border_color": "red"}}}
        },
        {
            "msg": "Pattern 6c: range with list of values or list of parameters and one format -"
                   "EXCEPTION: empty list of values",
            "test": {"C5:D@6": {"value": [], "format": {"border": 8, "border_color": "red"}}},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": "Pattern 6d: range with list of values or list of parameters and one format -"
                   "EXCEPTION: empty list of parameters",
            "test": {"C5:D@6": {"parameter": [], "format": {"border": 8, "border_color": "red"}}},
            "exception": "is not valid under any of the given schemas"
        },
        {
            "msg": "Pattern 6e: range with list of values or list of parameters and one format -"
                   "EXCEPTION: List of values, but empty format",
            "test": {"C5:D@6": {"value": ["pattern 6e", 2, 3], "format": {}}},
            "exception": "is not valid under any of the given schemas"
        },
    ]

    _zoom_config: dict = {
        "format": {
            "border": 7,
            "border_color": "grey"
        },
        "value": "--",
        "conditional_formattings": [
            {
                "range": "@A@1:@Z@9",
                "options": {
                    "type": "cell", "criteria": "equal to", "value": '"JSG"',
                    "format": {'bold': True, 'font_color': 'green'}
                }
            },
        ],
        # Pattern 1: single cell --> no list
        "@A@1": {
            "format": {"border": 8, "border_color": "red"},
            "value": "project ABC",
        },
        # Pattern 2: range which replicates one cell
        "@A@6:@E@6": {
            "format": {"border": 8, "border_color": "red"},
            "value": "***"
        },
        # Pattern 3: range as list of cells
        "@A@2:@A@5": [
            {
                "format": {"border": 8, "border_color": "red"},
                "value": "weight"
            },
            {
                "format": {"border": 8, "border_color": "red"},
                "value": "done"
            },
            {
                "format": {"border": 8, "border_color": "red"},
                "value": "progress"
            },
            {
                "format": {"border": 8, "border_color": "red"},
                "value": "velocity"
            },
        ],
        # Pattern 4: range with format and value as:
        # - Either one value to replicate (similar to pattern 2)
        # - Or list of values to apply (similar to patter 3)
        "@B@1:@D@1": {
            "format": {"border": 8, "border_color": "red"},
            "parameter": ["date_1", "date_2", "date_3"]
        },
        "@D@1:@D@1": {
            "format": [{"border": 8, "border_color": "red"}],
            "parameter": "date_1"
        },
        # Pattern 5: range to merge --> no list
        "@E@2:@E@5": {
            "value": "=100",
            "merged": True
        },
        # Check: default value
        "@F@1": {
            "format": [{"border": 8, "border_color": "red"}],
        },
        # Check: default value
        "@F@2": {
            "value": "jsg",
        },
    }
    _df_data: DataFrame = DataFrame(
        [
            [date(2022, 1, 1), 'project ABC', 100, 50],
            [date(2022, 8, 1), 'project ABC', 110, 75],
            [date(2022, 8, 28), 'project ABC', 100, 70],

            [date(2022, 1, 1), 'project DEF', 90, 45],
            [date(2022, 8, 1), 'project DEF', 100, 80],
            [date(2022, 8, 28), 'project DEF', 95, 81],
        ],
        columns=['measure date', 'project', 'weight', 'done']
    )

    def test_01_config_checker(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker: ConfigChecker = ConfigChecker(
            lst_widgets_json_schemas=[WidgetTableZoom.get_dict_config_schema()],
            str_base_json_schema_id=WidgetTableZoom.get_widget_id()
        )

        for i_test in self._test_data_by_pattern:
            if "exception" in i_test:
                with self.assertRaisesRegex(
                    Exception,
                    i_test["exception"],
                    msg=f"Exception not raised when executing test '{i_test['msg']}'."
                ):
                    _the_config_checker.validate(
                        dict_to_validate=i_test['test'],
                    )
            else:
                try:
                    _the_config_checker.validate(
                        dict_to_validate=i_test['test'],
                    )
                except ValidationError as exception:
                    raise Exception(f"Unexpected exception when executing test '{i_test['msg']}'.") from exception

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_02_unitary(self):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We instantiate the widget
        _the_widget: WidgetTableZoom = WidgetTableZoom(
            str_address="B2",
            dict_parameters_map={},
            dict_config=self._zoom_config,
            dict_raw_parameters={
                "date_3": "that is date 3",
                "date_1": "that is date 1",
                "Trend": "that is a trend",
                "date_2": "that is date 2"
            },
            str_sheet_name="abc",
            dict_default_format={}
        )

        # We print it into an excel spreadsheet
        _the_workbook = xlsxwriter.Workbook(Path(os.path.join(os.path.dirname(__file__), 'TC02.excel.xlsx')))
        _the_worksheet = _the_workbook.add_worksheet()

        # We print the widget
        _the_widget.write_to_excel(
            wb_workbook=_the_workbook,
            ws_worksheet=_the_worksheet
        )

        # We close the workbook
        _the_workbook.close()

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function_list_parameters(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _lst_the_parameters: list = WidgetTableZoom._list_parameters(
            config=self._zoom_config
        )

        self.assertListEqual(
            list1=['date_1', 'date_2', 'date_3'],
            list2=_lst_the_parameters,
            msg=f"Function list_parameters did not return expected results"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function_get_parameters(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _dict_the_parameters: dict = WidgetTableZoom._get_parameters(
            config=self._zoom_config,
            dict_parameters={"date_2": date(2022, 9, 3), "date_3": "", "Trend": "", "date_1": "1"}
        )

        self.assertDictEqual(
            d1={"date_2": date(2022, 9, 3), "date_3": "", "date_1": "1"},
            d2=_dict_the_parameters,
            msg=f"Function get_parameters did not return expected results"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function_process_range(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _list_the_test_data: list = [
            {
                "msg": "# Pattern 1: single cell --> no list",
                "test": {
                    "C5": {"format": {"border": 8, "border_color": "red"}, "value": "project ABC"}
                },
                "expected": {
                    'C5': {
                        'function': 'write',
                        'parameter': {"format": {"border": 8, "border_color": "red"}, "value": "project ABC"}
                    }
                },
            },
            {
                "msg": "# Pattern 2: range which replicates one cell",
                "test": {"@A@6:@B@6": {"parameter": "param_1"}},
                "expected": {
                    'B8': {'function': 'write', 'parameter': {"format": {"border": 1}, "value": "val_1"}},
                    'C8': {'function': 'write', 'parameter': {"format": {"border": 1}, "value": "val_1"}},
                },
            },
            {
                "msg": "# Pattern 3: range as list of cells",
                "test": {
                    "@A@6:@B@6": [
                        {"format": {"border": 2}},
                        {"value": "pattern 3"}
                    ]
                },
                "expected": {
                    'B8': {'function': 'write', 'parameter': {"format": {"border": 2}, "value": "default"}},
                    'C8': {'function': 'write', 'parameter': {"format": {"border": 1}, "value": "pattern 3"}},
                },
            },
            {
                "msg": "# Pattern 4a: range with format and value as: (i) either one value to replicate (similar to "
                       "pattern 2) or (ii) list of values to apply (similar to patter 3)",
                "test": {
                    "@B@1:@D@1": {
                        "format": {"border_color": "brown"},
                        "parameter": ["param_1", "param_2", "param_3"]
                    }
                },
                "expected": {
                    'C3': {'function': 'write', 'parameter': {"format": {"border_color": "brown"}, "value": "val_1"}},
                    'D3': {'function': 'write', 'parameter': {"format": {"border_color": "brown"},
                                                              "value": date(2022, 9, 9)}},
                    'E3': {'function': 'write', 'parameter': {"format": {"border_color": "brown"}, "value": 1.5}},
                },
            },
            {
                "msg": "# Pattern 4b: range with format and value as: (i) either one value to replicate (similar to "
                       "pattern 2) or (ii) list of values to apply (similar to patter 3)",
                "test": {
                    "@D@1:@E@1": {
                        "format": [{"border": 0}, {"border": 1}],
                        "parameter": "param_1"
                    },
                },
                "expected": {
                    'E3': {'function': 'write', 'parameter': {"format": {"border": 0}, "value": "val_1"}},
                    'F3': {'function': 'write', 'parameter': {"format": {"border": 1}, "value": "val_1"}},
                },
            },
            {
                "msg": "# Pattern 5: range to merge --> no list",
                "test": {
                    "@E@2:@E@$5": {
                        "value": "=100",
                        "merged": True
                    }
                },
                "expected": {
                    'F4:F7': {'function': 'merge_range', 'parameter': {"format": {"border": 1}, "value": "=100"}}
                },
            },
        ]

        for i_test in _list_the_test_data:

            i_test_case: list = list(i_test["test"].items())[0]

            _the_result: dict = WidgetTableZoom._process_range(
                str_reference_address="B3",
                str_address=i_test_case[0],
                config=i_test_case[1],
                dict_parameters={"param_1": "val_1", "param_2": date(2022, 9, 9), "param_3": 1.5},
                dict_default_format={"border": 1},
                default_value="default"
            )

            self._logger.info(f"-------------------------------------------")
            self._logger.info(f"TEST CASE: {i_test['msg']}")
            self._logger.info(f"RESULT   : \n{_the_result}")

            self.assertDictEqual(
                d1=i_test["expected"],
                d2=_the_result,
                msg=i_test["msg"]
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function_for_each_cell_in_range(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _lst_the_test_data: list = [
            {
                "msg": "not a range but one cell",
                "test": "A1",
                "expected": ["A1"]
            },
            {
                "msg": "not a range but one cell",
                "test": "$A1",
                "expected": ["A1"]
            },
            {
                "msg": "a simple range",
                "test": "A1:C2",
                "expected": ['A1', 'B1', 'C1', 'A2', 'B2', 'C2']
            },
            {
                "msg": "a simple range",
                "test": "$A1:C$2",
                "expected": ['A1', 'B1', 'C1', 'A2', 'B2', 'C2']
            },
        ]

        for i_test in _lst_the_test_data:
            self.assertListEqual(
                list1=i_test["expected"],
                list2=WidgetTableZoom._for_each_cell_in_range(i_test["test"]),
                msg=i_test["msg"]
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import unittest
from inspect import stack
from logging import getLogger, Logger

from com_enovation.toolbox.excel_dashboard.cell_notation_helper import CellNotationHelper


class TestFunction__IsAddressXxx(unittest.TestCase):

    _logger: Logger = getLogger(__name__)

    _list_the_incorrect_addresses: list[str] = [
        " D3",
        "32D",
        "3D",
        "A02"
    ]

    _list_the_cell_addresses: list[str] = [
        "C5",

        "@C5",
        "@$C5",
        "$C5",

        "C@5",
        "C@$5",
        "C$5",

        "@$C@$5",
        "@C$5",
        "$C@5",
    ]

    _list_the_range_addresses: list[str] = [
        "C5:@CA61",

        "@C5:@$CA61",
        "@$C5:$CA61",
        "$C5:CA@61",

        "C@5:CA@$61",
        "C@$5:CA$61",
        "C$5:@$CA@$61",

        "@$C@$5:@CA$61",
        "@C$5:$CA@61",
        "$C@5:CA61",
    ]

    _list_the_column_addresses: list[str] = [
        "C:CA",

        "@C:CA",
        "@$C:CA",
        "$C:CA",

        "C:@CA",
        "C:@$CA",
        "C:$CA",

        "@$C:@$CA",
        "$C:@CA",
    ]

    _list_the_row_addresses: list[str] = [
        "5:63",

        "@5:63",
        "@$5:63",
        "$5:63",

        "5:@63",
        "5:@$63",
        "5:$63",

        "@5:@$63",
        "$5:@63",
    ]

    def test_function__is_address_cell_pattern(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_str_address in self._list_the_cell_addresses:
            self.assertTrue(
                CellNotationHelper.is_address_cell_pattern(str_address=i_str_address),
                msg=f"Error when process address '{i_str_address}', expected to be OK."
            )

        for i_str_address in \
                self._list_the_range_addresses + \
                self._list_the_column_addresses + \
                self._list_the_row_addresses:
            self.assertFalse(
                CellNotationHelper.is_address_cell_pattern(str_address=i_str_address),
                msg=f"Error when process address '{i_str_address}', expected to be NOK."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__is_address_range_pattern(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_str_address in self._list_the_range_addresses:
            self.assertTrue(
                CellNotationHelper.is_address_range_pattern(str_address=i_str_address),
                msg=f"Error when process address '{i_str_address}', expected to be OK."
            )

        for i_str_address in \
                self._list_the_cell_addresses + \
                self._list_the_column_addresses + \
                self._list_the_row_addresses:
            self.assertFalse(
                CellNotationHelper.is_address_range_pattern(str_address=i_str_address),
                msg=f"Error when process address '{i_str_address}', expected to be NOK."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__is_address_column_pattern(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_str_address in self._list_the_column_addresses:
            self.assertTrue(
                CellNotationHelper.is_address_column_pattern(str_address=i_str_address),
                msg=f"Error when process address '{i_str_address}', expected to be OK."
            )

        for i_str_address in \
                self._list_the_cell_addresses + \
                self._list_the_range_addresses + \
                self._list_the_row_addresses:
            self.assertFalse(
                CellNotationHelper.is_address_column_pattern(str_address=i_str_address),
                msg=f"Error when process address '{i_str_address}', expected to be NOK."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__is_address_row_pattern(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_str_address in self._list_the_row_addresses:
            self.assertTrue(
                CellNotationHelper.is_address_row_pattern(str_address=i_str_address),
                msg=f"Error when process address '{i_str_address}', expected to be OK."
            )

        for i_str_address in \
                self._list_the_cell_addresses + \
                self._list_the_range_addresses + \
                self._list_the_column_addresses:
            self.assertFalse(
                CellNotationHelper.is_address_row_pattern(str_address=i_str_address),
                msg=f"Error when process address '{i_str_address}', expected to be NOK."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__is_address_pattern_correct(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_str_address in \
                self._list_the_cell_addresses + \
                self._list_the_range_addresses + \
                self._list_the_column_addresses + \
                self._list_the_row_addresses:
            self.assertTrue(
                CellNotationHelper.is_address_pattern_correct(str_address=i_str_address),
                msg=f"Error when process address '{i_str_address}', expected to be OK."
            )

        for i_str_address in self._list_the_incorrect_addresses:
            self.assertFalse(
                CellNotationHelper.is_address_pattern_correct(str_address=i_str_address),
                msg=f"Error when process address '{i_str_address}', expected to be NOK."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_exception(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_str_address in self._list_the_incorrect_addresses:
            with self.assertRaisesRegex(Exception, f"The address '{i_str_address}' is not recognized as an address, "
                                                   f"and its pattern could not be determined."):
                CellNotationHelper.is_address_cell_pattern(str_address=i_str_address),

            with self.assertRaisesRegex(Exception, f"The address '{i_str_address}' is not recognized as an address, "
                                                   f"and its pattern could not be determined."):
                CellNotationHelper.is_address_range_pattern(str_address=i_str_address),

            with self.assertRaisesRegex(Exception, f"The address '{i_str_address}' is not recognized as an address, "
                                                   f"and its pattern could not be determined."):
                CellNotationHelper.is_address_column_pattern(str_address=i_str_address),

            with self.assertRaisesRegex(Exception, f"The address '{i_str_address}' is not recognized as an address, "
                                                   f"and its pattern could not be determined."):
                CellNotationHelper.is_address_row_pattern(str_address=i_str_address),

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import unittest
from inspect import stack
from logging import getLogger, Logger

from com_enovation.toolbox.excel_dashboard.cell_notation_helper import CellNotationHelper


class test_function__get_address_pattern(unittest.TestCase):

    _logger: Logger = getLogger(__name__)

    def test_function(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _dict_the_test_data: dict = {

            # Pattern 1, cells
            "D3": "--A--1",

            "@D3": "@-A--1",
            "@$D3": "@$A--1",
            "$D3": "-$A--1",

            "D@3": "--A@-1",
            "D@$3": "--A@$1",
            "D$3": "--A-$1",

            "@$D@$3": "@$A@$1",

            # Pattern 2, range
            "C2:C3": "--A--1:--A--1",
            "C@2:C$3": "--A@-1:--A-$1",

            # Pattern 3, column
            "C:C": "--A:--A",
            "@C:@$C": "@-A:@$A",

            # Pattern 4, row
            "@$3:$3": "@$1:-$1"

        }

        for k, v in _dict_the_test_data.items():
            self.assertEqual(
                v,
                CellNotationHelper.get_address_pattern(
                    str_address=k
                ),
                msg=f"Error when trying to get pattern for '{k}'."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_exception(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _dict_the_test_data: list[str] = [
            " D3",
            "32D"
        ]

        for i_address in _dict_the_test_data:

            with self.assertRaisesRegex(
                    Exception,
                    f"The address '{i_address}' is not recognized as an address, and its pattern could not "
                    f"be determined."
            ):
                CellNotationHelper.get_address_pattern(
                    str_address=i_address
                ),

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from inspect import stack
from unittest import TestCase
from logging import Logger, getLogger

from pandas import DataFrame

from com_enovation.helper.pandas_dataframe_typer import PandasDataframeTyper
from com_enovation.toolbox.backlog.bck_compare.bck_comparer import BacklogComparer


class TestFunctionCheckParameterDictColumns(TestCase):
    _logger: Logger = getLogger(__name__)

    def test_001(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        BacklogComparer()._check_parameter_dict_columns(
            df_src=DataFrame(
                {
                    "key1": ["k1-valA", "k1-valB"],
                    "col1": ["col1-valA", "col1-valB"]
                }
            ),
            df_tgt=DataFrame(
                {
                    "key1": ["k1-valA", "k1-valB"],
                    "col1": ["col1-valA", "col1-valB"]
                }
            ),
            dict_columns={
                BacklogComparer.str__dict_columns__json_tag__key_columns: {
                    "key1": {
                        BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                    }
                },
                BacklogComparer.str__dict_columns__json_tag__compared_columns: {
                    "col1": {
                        BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                    }
                }
            }
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_BR_101(self):
        """
        BR_101: parameter "key columns" should exist
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        with self.assertRaisesRegex(
                KeyError,
                f"'key columns'"
        ):

            BacklogComparer()._check_parameter_dict_columns(
                df_src=DataFrame(
                    {
                        "key1": ["k1-valA", "k1-valB"],
                        "col1": ["col1-valA", "col1-valB"]
                    }
                ),
                df_tgt=DataFrame(
                    {
                        "key1": ["k1-valA", "k1-valB"],
                        "col1": ["col1-valA", "col1-valB"]
                    }
                ),
                dict_columns={
                    # BacklogComparer.str__dict_columns__json_tag__key_columns: {
                    #     "key1": {
                    #         BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                    #     }
                    # },
                    BacklogComparer.str__dict_columns__json_tag__compared_columns: {
                        "col1": {
                            BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                        }
                    }
                }
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_BR_102(self):
        """
        BR_102: parameter "key columns" should be a dictionary
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        with self.assertRaisesRegex(
                AttributeError,
                f"'str' object has no attribute 'keys'"
        ):
            BacklogComparer()._check_parameter_dict_columns(
                df_src=DataFrame(
                    {
                        "key1": ["k1-valA", "k1-valB"],
                        "col1": ["col1-valA", "col1-valB"]
                    }
                ),
                df_tgt=DataFrame(
                    {
                        "key1": ["k1-valA", "k1-valB"],
                        "col1": ["col1-valA", "col1-valB"]
                    }
                ),
                dict_columns={
                    BacklogComparer.str__dict_columns__json_tag__key_columns: "key 1",
                    BacklogComparer.str__dict_columns__json_tag__compared_columns: {
                        "col1": {
                            BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                        }
                    }
                }
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_BR_103(self):
        """
        BR_103: parameter "key columns" should contain at least one property
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        with self.assertRaisesRegex(
                Exception,
                f"^"
                f"{{}} does not have enough properties"
                f"[\\s,\\S]*"
                f"On instance\\['{BacklogComparer.str__dict_columns__json_tag__key_columns}'\\]"
                f"[\\S,\\s]*"
                f"$"
        ):
            BacklogComparer()._check_parameter_dict_columns(
                df_src=DataFrame(
                    {
                        "key1": ["k1-valA", "k1-valB"],
                        "col1": ["col1-valA", "col1-valB"]
                    }
                ),
                df_tgt=DataFrame(
                    {
                        "key1": ["k1-valA", "k1-valB"],
                        "col1": ["col1-valA", "col1-valB"]
                    }
                ),
                dict_columns={
                    BacklogComparer.str__dict_columns__json_tag__key_columns: {
                        # "key1": {
                        #     BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                        # }
                    },
                    BacklogComparer.str__dict_columns__json_tag__compared_columns: {
                        "col1": {
                            BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                        }
                    }
                }
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_BR_104(self):
        """
        BR_104: parameter "key columns" should only contains keys among columns intersecting in both dataframes
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        with self.assertRaisesRegex(
                Exception,
                f"Additional properties are not allowed \\('key2' was unexpected\\)"
        ):
            BacklogComparer()._check_parameter_dict_columns(
                df_src=DataFrame(
                    {
                        "key1": ["k1-valA", "k1-valB"],
                        "col1": ["col1-valA", "col1-valB"]
                    }
                ),
                df_tgt=DataFrame(
                    {
                        "key1": ["k1-valA", "k1-valB"],
                        "key2": ["k1-valA", "k1-valB"],
                        "col1": ["col1-valA", "col1-valB"]
                    }
                ),
                dict_columns={
                    BacklogComparer.str__dict_columns__json_tag__key_columns: {
                        "key2": {
                            BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                        }
                    },
                    BacklogComparer.str__dict_columns__json_tag__compared_columns: {
                        "col1": {
                            BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                        }
                    }
                }
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_BR_105(self):
        """
        BR_105: parameter "key columns" should contains records containing "type" key only
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        with self.assertRaisesRegex(
                Exception,
                f"Additional properties are not allowed \\('jsg' was unexpected\\)"
        ):
            BacklogComparer()._check_parameter_dict_columns(
                df_src=DataFrame(
                    {
                        "key1": ["k1-valA", "k1-valB"],
                        "col1": ["col1-valA", "col1-valB"]
                    }
                ),
                df_tgt=DataFrame(
                    {
                        "key1": ["k1-valA", "k1-valB"],
                        "col1": ["col1-valA", "col1-valB"]
                    }
                ),
                dict_columns={
                    BacklogComparer.str__dict_columns__json_tag__key_columns: {
                        "key1": {
                            BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                            "jsg": "unexpected",
                        }
                    },
                    BacklogComparer.str__dict_columns__json_tag__compared_columns: {
                        "col1": {
                            BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                        }
                    }
                }
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_BR_106(self):
        """
        BR_106: parameter "key columns" should contains records containing "type" key
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        with self.assertRaisesRegex(
                Exception,
                f"'type' is a required property"
        ):
            BacklogComparer()._check_parameter_dict_columns(
                df_src=DataFrame(
                    {
                        "key1": ["k1-valA", "k1-valB"],
                        "col1": ["col1-valA", "col1-valB"]
                    }
                ),
                df_tgt=DataFrame(
                    {
                        "key1": ["k1-valA", "k1-valB"],
                        "col1": ["col1-valA", "col1-valB"]
                    }
                ),
                dict_columns={
                    BacklogComparer.str__dict_columns__json_tag__key_columns: {
                        "key1": {
                            # BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                        }
                    },
                    BacklogComparer.str__dict_columns__json_tag__compared_columns: {
                        "col1": {
                            BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                        }
                    }
                }
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_BR_107(self):
        """
        BR_107: parameter "key columns - type" should be a string
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        with self.assertRaisesRegex(
                Exception,
                f"'dummy' is not one of \\['date', 'float', 'int', 'str'\\]"
        ):
            BacklogComparer()._check_parameter_dict_columns(
                df_src=DataFrame(
                    {
                        "key1": ["k1-valA", "k1-valB"],
                        "col1": ["col1-valA", "col1-valB"]
                    }
                ),
                df_tgt=DataFrame(
                    {
                        "key1": ["k1-valA", "k1-valB"],
                        "col1": ["col1-valA", "col1-valB"]
                    }
                ),
                dict_columns={
                    BacklogComparer.str__dict_columns__json_tag__key_columns: {
                        "key1": {
                            BacklogComparer.str__dict_columns__json_tag__type: "dummy",
                        }
                    },
                    BacklogComparer.str__dict_columns__json_tag__compared_columns: {
                        "col1": {
                            BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                        }
                    }
                }
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import json
import os
import unittest
from pathlib import Path
from logging import Logger, getLogger
from inspect import stack
from pandas import DataFrame, read_excel

from com_enovation.toolbox.data_handler.data_handler import DataHandler


class TestDataHandler(unittest.TestCase):
    _logger: Logger = getLogger(__name__)

    def test_data_handler(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the config
        with open(Path(os.path.join(os.path.dirname(__file__), '01.config.json')), 'r') as json_file:
            dict_config: dict = json.load(json_file)

        # we instantiate the data handler
        _dh_the_data_handler: DataHandler = DataHandler(dict_config=dict_config)

        # We load the data to handle
        _df_the_data_extract: DataFrame = read_excel(
            Path(os.path.join(os.path.dirname(__file__), '02.data_to_handle.xlsx'))
        )

        # We handle the data: one single sequence
        _df_the_return: DataFrame = _dh_the_data_handler.handle_data(
            df_data=_df_the_data_extract,
            sequence_alias="first sequence"
        )

        # We handle the data: one single sequence, but as a list
        _df_the_return: DataFrame = _dh_the_data_handler.handle_data(
            df_data=_df_the_data_extract,
            sequence_alias=["first sequence", "second sequence"]
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import logging
from inspect import stack

import click
from pandas import DataFrame, read_csv, read_excel

from com_enovation.helper.click.python_literal_argument_and_option import PythonLiteralOption

_logger: logging.Logger = logging.getLogger(__name__)


@click.command('df-load-csv')
@click.pass_context
@click.argument('file', type=click.Path(exists=True))
@click.argument('alias', type=str, default='csv')
@click.option('-c', '--columns', type=list, cls=PythonLiteralOption, default=[],
              help="Columns labels to load, others will be discarded.")
def df_load_csv(ctx_context, file, columns, alias):
    """
    Load listed columns from a csv file into a dataframe that is labelled using an alias for later use.

    :param alias: name given to the dataframe being loaded, so it can be used later on by other commands.
    :param columns: columns from the CSV file to be loaded. Other columns will be ignored.
    :param ctx_context: Click context.
    :param file: csv file to load
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias in ctx_context.obj:
        _logger.warning(
            f"Warning: another object with alias '{alias}' already exists, of type {type(alias)}. This data "
            f"will be overridden with the 'csv' file '{file}'."
        )

    # We load the csv
    df_the_measures: DataFrame = read_csv(file)
    _logger.info(f"CSV file read, shape '{df_the_measures.shape}'.")

    # We only keep the listed columns, if such "listed columns" were even provided as parameter
    if columns:

        # We check the listed columns are effectively in the data source
        lst_the_inexistant_columns: list = list(set(columns) - set(df_the_measures.columns))
        if len(lst_the_inexistant_columns) > 0:
            raise Exception(
                f"selected columns '{', '.join(lst_the_inexistant_columns)}' are not among existing ones "
                f"'{', '.join(df_the_measures.columns)}'."
            )

        # We filter out non listed columns
        df_the_measures = df_the_measures[list(columns)]
        _logger.info(f"Dataframe reduced, shape '{df_the_measures.shape}'.")

    # We update the context data store
    ctx_context.obj[alias] = df_the_measures
    ctx_context.obj["_" + alias] = {
        "path": file,
        "src": "load_csv"
    }

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")


@click.command('df-load-xls')
@click.pass_context
@click.argument('file', type=click.Path(exists=True))
@click.argument('alias', type=str, default='xls')
@click.option('-c', '--columns', type=list, cls=PythonLiteralOption, default=[],
              help="Columns labels to load, others will be discarded.")
@click.option('-k', '--kwargs', type=dict, cls=PythonLiteralOption, default={},
              help="Parameters for the underlying function pandas.read_excel.")
def df_load_xls(ctx_context, file, columns, alias, kwargs):
    """
    Load an Excel spreadsheet from path FILE, and store it in context as ALIAS.

    In case COLUMNS is provided, only the listed columns are retrieved.

    In case KWARGS is provided, it is provided as parameters to the underlying function pandas.read_excel.
    Illustration: enov --verbose df-load-xls --kwargs "{'sheet_name': 'whatever'}" <PATH>
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

    if alias in ctx_context.obj:
        _logger.warning(
            f"Warning: another object with alias '{alias}' already exists, of type {type(alias)}. This data "
            f"will be overridden with the 'xls-x' file '{file}'."
        )

    # We load the csv
    if len(kwargs) == 0:
        df_the_measures: DataFrame = read_excel(file)
    else:
        df_the_measures: DataFrame = read_excel(file, **kwargs)
    _logger.info(f"XLS-x file read, shape '{df_the_measures.shape}'.")

    # We only keep the listed columns, if such "listed columns" were even provided as parameter
    if columns:

        # We check the listed columns are effectively in the data source
        lst_the_inexistant_columns: list = list(set(columns) - set(df_the_measures.columns))
        if len(lst_the_inexistant_columns) > 0:
            raise Exception(
                f"selected columns '{', '.join(lst_the_inexistant_columns)}' are not among existing ones "
                f"'{', '.join(df_the_measures.columns)}'."
            )

        # We filter out non listed columns
        df_the_measures = df_the_measures[list(columns)]
        _logger.info(f"Dataframe reduced, shape '{df_the_measures.shape}'.")

    _logger.info(f"Stored as '{alias}' in context.")

    # We update the context data store
    ctx_context.obj[alias] = df_the_measures
    ctx_context.obj["_" + alias] = {
        "path": file,
        "src": "load_xls"
    }

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from inspect import stack
from logging import Logger, getLogger
from unittest import TestCase

from jsonschema import ValidationError

from com_enovation.toolbox.excel_dashboard.config_checker import ConfigChecker


class TestAddressRange(TestCase):

    _logger: Logger = getLogger(__name__)

    def test_01_validated(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/address/range",
            lst_widgets_json_schemas=[]
        )

        for i_range in [
            "A1",
            "$A$1",
            "A1 A2",
            "A1 B2:$D$34"
        ]:
            # noinspection PyTypeChecker
            _the_config_checker.validate(dict_to_validate=i_range)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_02_not_validated(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/address/range",
            lst_widgets_json_schemas=[]
        )

        for i_range in [
            "A1q",
            "$A$1 ",
            "A1  A2",
            "A1 2:$D$34"
        ]:
            with self.assertRaisesRegex(
                    ValidationError,
                    f"does not match"
            ):
                # noinspection PyTypeChecker
                _the_config_checker.validate(dict_to_validate=i_range)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import json
import os
import unittest
from inspect import stack
from logging import Logger, getLogger
from pathlib import Path

from com_enovation.toolbox.data_broadcaster.data_broadcaster import DataBroadcaster


class TestDataBroadcaster(unittest.TestCase):
    """
    In this test, we will do execute the following logic:
    - We load the various data (data prepared ahead of calling the broadcaster)
    - We then broadcast these data into several dashboards, after having filtered (and possibly refreshed) the data
    """
    _logger: Logger = getLogger(__name__)

    def test_data_broadcaster(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the config
        with open(Path(os.path.join(os.path.dirname(__file__), 'IN_test_data/BroadcasterConfiguration.json')), 'r') \
                as json_file:
            dict_config: dict = json.load(json_file)

        # From this distribution configuration, we adjust the paths to the Excel dashboards configurations
        for k_dashboard_config, v_dashboard_config in \
                dict_config[DataBroadcaster._const_str__config_tag__dashboards_cfgs__as_dict].items():
            v_dashboard_config[DataBroadcaster._const_str__config_tag__dashboard_cfg__path__as_path] = Path(
                os.path.join(
                    os.path.dirname(__file__),
                    "IN_test_data",
                    v_dashboard_config[DataBroadcaster._const_str__config_tag__dashboard_cfg__path__as_path]
                )
            )

        _obj_the_broadcaster: DataBroadcaster = DataBroadcaster(
            dict_config=dict_config
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import logging
from pathlib import Path

import click
from click import Context

from com_enovation.helper.excel.xls_data_connection_to_df import XlsDataConnectionToDf, ExcelTables

_logger: logging.Logger = logging.getLogger(__name__)


@click.command('xls-connection-to-df')
@click.argument('file', type=click.Path(exists=True))
@click.argument('alias', type=str, default='xls')
@click.option('--persist/--no-persist', default=True, help='To persist the refresh excel file.')
@click.pass_context
def xls_connection_to_df(
        ctx_context: Context,
        file: Path,
        alias: str,
        persist: bool
):
    """
    Command to open an Excel file from PATH, refresh all queries and connections, before persisting the file (if
    PERSIST is left to true) and returning all the tables in an ExcelTables bean. This bean is persisted in context as
    ALIAS (defaulted to 'xls').

    Note: when crashing, a zombie process might survive:
    - ps | grep EXCEL
    - kill xxx
    """

    if alias in ctx_context.obj:
        _logger.warning(
            f"Warning: another object with alias '{alias}' already exists, of type {type(alias)}. This data "
            f"will be overridden."
        )

    obj_bean: ExcelTables = XlsDataConnectionToDf().refresh_and_return_xls_connections(
        p_path=Path(file),
        b_persist=persist
    )

    # We update the context data store
    ctx_context.obj[alias] = obj_bean
    ctx_context.obj["_" + alias] = {
        "path": file,
        "src": "xls-connections-to-df"
    }



from unittest import TestCase
from pandas import DataFrame
from inspect import stack
from logging import Logger, getLogger
from datetime import datetime

from pandas.testing import assert_frame_equal

from com_enovation.toolbox.predictability.dp_date_predictability.dp_computer import DatePredictabilityComputer


class TestFunctionCheckAndCleanseData(TestCase):

    _logger: Logger = getLogger(__name__)

    def test_BR_Data(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        the_df_client_measures: DataFrame = DataFrame(columns=["key", "a date column", "measure"], data=[
            ['a', datetime(2020, 5, 10), datetime(2020, 5, 18)],
            ['a', datetime(2020, 5, 19), datetime(2020, 5, 19)],
        ])
        DatePredictabilityComputer()._check_and_cleanse_data(
            df_measures=the_df_client_measures,
            str_column_label_date="a date column",
        )

        self.assertEqual(
            first=2,
            second=len(
                DatePredictabilityComputer()._check_and_cleanse_data(
                    df_measures=the_df_client_measures,
                    str_column_label_date="a date column",
                ).index
            ),
            msg="Unexpected number of lines in returned DataFrame."
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_LOG_Data_002_Remove_duplicates(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # 1. Duplicate from first record --> one line deleted
        the_df_client_measures: DataFrame = DataFrame(columns=["key", "date", "measure"], data=[
            ['a', datetime(2020, 5, 10), datetime(2020, 5, 18)],
            ['a', datetime(2020, 5, 17), datetime(2020, 5, 18)],
            ['a', datetime(2020, 5, 19), datetime(2020, 5, 19)],
        ])
        assert_frame_equal(
            left=DataFrame(columns=["key", "date", "measure"], data=[
                ['a', datetime(2020, 5, 10), datetime(2020, 5, 18)],
                ['a', datetime(2020, 5, 19), datetime(2020, 5, 19)],
            ]),
            right=DatePredictabilityComputer()._check_and_cleanse_data(
                df_measures=the_df_client_measures
            ).reset_index(drop=True),
            obj=f"UC 1, did not delete properly duplicates",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        # 2. Duplicate from last record --> no line deleted
        the_df_client_measures: DataFrame = DataFrame(columns=["key", "date", "measure"], data=[
            ['a', datetime(2020, 5, 10), datetime(2020, 5, 18)],
            ['a', datetime(2020, 5, 17), datetime(2020, 5, 19)],
            ['a', datetime(2020, 5, 19), datetime(2020, 5, 19)],
        ])
        assert_frame_equal(
            left=the_df_client_measures,
            right=DatePredictabilityComputer()._check_and_cleanse_data(
                df_measures=the_df_client_measures
            ).reset_index(drop=True),
            obj=f"UC 2, did not delete properly duplicates",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        # 3. Duplicate in the middle --> several lines deleted
        the_df_client_measures: DataFrame = DataFrame(columns=["key", "date", "measure"], data=[
            ['a', datetime(2020, 5, 10), datetime(2020, 5, 18)],
            ['a', datetime(2020, 5, 11), datetime(2020, 5, 18)],
            ['a', datetime(2020, 5, 12), datetime(2020, 5, 18)],
            ['a', datetime(2020, 5, 13), datetime(2020, 5, 19)],
            ['a', datetime(2020, 5, 14), datetime(2020, 5, 19)],
            ['a', datetime(2020, 5, 15), datetime(2020, 5, 19)],
            ['a', datetime(2020, 5, 16), datetime(2020, 5, 20)],
            ['a', datetime(2020, 5, 17), datetime(2020, 5, 20)],
            ['a', datetime(2020, 5, 18), datetime(2020, 5, 19)],
            ['a', datetime(2020, 5, 19), datetime(2020, 5, 19)],
        ])
        assert_frame_equal(
            left=DataFrame(columns=["key", "date", "measure"], data=[
                ['a', datetime(2020, 5, 10), datetime(2020, 5, 18)],
                ['a', datetime(2020, 5, 13), datetime(2020, 5, 19)],
                ['a', datetime(2020, 5, 16), datetime(2020, 5, 20)],
                ['a', datetime(2020, 5, 18), datetime(2020, 5, 19)],
                ['a', datetime(2020, 5, 19), datetime(2020, 5, 19)],
            ]),
            right=DatePredictabilityComputer()._check_and_cleanse_data(
                df_measures=the_df_client_measures
            ).reset_index(drop=True),
            obj=f"UC 2, did not delete properly duplicates",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_BR_Data_001_at_least_one_record(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # 1. Exception raised with no record
        the_df_client_measures: DataFrame = DataFrame(columns=["a key", "a date column", "measure"], data=[])
        with self.assertRaisesRegex(
                Exception,
                f"BR_Data_001 - The parameter df_measures contains '0' records, while it should contain at least '1'."
        ):
            DatePredictabilityComputer()._check_and_cleanse_data(
                df_measures=the_df_client_measures,
                str_column_label_key="a key",
                str_column_label_date="a date column"
            )

        # 2. No exception raised with one record
        the_df_client_measures: DataFrame = DataFrame(columns=["key", "a date column", "measure"], data=[
            ["a", datetime(2020, 5, 17), datetime(2020, 5, 18)],
        ])
        DatePredictabilityComputer()._check_and_cleanse_data(
            df_measures=the_df_client_measures,
            str_column_label_date="a date column",
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_BR_Data_002_no_null_value(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        the_df_client_measures: DataFrame = DataFrame(columns=["date", "measure"], data=[
            [datetime(2020, 5, 17), datetime(2020, 5, 18)],
            [datetime(2020, 5, 18), None],
        ])

        with self.assertRaisesRegex(
                Exception,
                f"BR_Data_002 - The parameter df_measures contains '1' null values, which is not expected. Make sure "
                f"you call the function with a dataframe df_measures which does not contain any null value."
        ):
            DatePredictabilityComputer()._check_and_cleanse_data(
                df_measures=the_df_client_measures,
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_BR_Data_003_dates(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # 1. No exception when string can be formatted as date
        the_df_client_measures: DataFrame = DataFrame(columns=["key", "date", "measure"], data=[
            ['a', "2020-5-17", datetime(2020, 5, 18)],
            ['a', "2020-5-18", datetime(2020, 5, 18)],
        ])
        DatePredictabilityComputer()._check_and_cleanse_data(
            df_measures=the_df_client_measures,
        )

        # 2. Exception when string cannot be formatted as date
        the_df_client_measures: DataFrame = DataFrame(columns=["key", "date", "measure"], data=[
            ['a', "2020-5-aa", datetime(2020, 5, 18)],
            ['a', "2020-5-18", datetime(2020, 5, 18)],
        ])
        self._logger.warning(f"The following warning is to be expected! UserWarning: Could not infer format, so each "
                           f"element will be parsed individually, falling back to `dateutil`. To ensure parsing is "
                           f"consistent and as-expected, please specify a format\n"
                           f"... START")
        with self.assertRaisesRegex(
                Exception,
                f"BR_Data_003 - An exception occurred while typing columns 'date' and 'measure' to dates. Check the "
                f"data, and retry."
        ):
            DatePredictabilityComputer()._check_and_cleanse_data(
                df_measures=the_df_client_measures,
            )
        self._logger.warning(f"... END")

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_BR_Data_004_duplicated_date(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # 1. duplicated date without key column
        the_df_client_measures: DataFrame = DataFrame(columns=["date", "measure"], data=[
            ["2020-05-18", datetime(2020, 5, 18)],
            ["2020-05-18", datetime(2020, 5, 18)],
        ])
        with self.assertRaisesRegex(
                Exception,
                f"BR_Data_004 - The parameter df_measures contains '1' records duplicated, which is not expected. Make "
                f"sure your remove all duplicates for values"
        ):
            DatePredictabilityComputer()._check_and_cleanse_data(
                df_measures=the_df_client_measures,
            )

        # 2. duplicated date with one key column
        the_df_client_measures: DataFrame = DataFrame(columns=["key", "date", "measure"], data=[
            ["a key", "2020-05-18", datetime(2020, 5, 18)],
            ["a key", "2020-05-18", datetime(2020, 5, 18)],
        ])
        with self.assertRaisesRegex(
                Exception,
                f"BR_Data_004 - The parameter df_measures contains '1' records duplicated, which is not expected. Make "
                f"sure your remove all duplicates for values"
        ):
            DatePredictabilityComputer()._check_and_cleanse_data(
                df_measures=the_df_client_measures,
            )

        # 3. duplicated date with one several key columns
        the_df_client_measures: DataFrame = DataFrame(columns=["key 1", "key 2", "date", "measure"], data=[
            ["a", "b", "2020-05-17", datetime(2020, 5, 18)],
            ["a", "b", "2020-05-18", datetime(2020, 5, 18)],
            ["a", "c", "2020-05-18", datetime(2020, 5, 18)],
            ["a", "c", "2020-05-18", datetime(2020, 5, 18)],
        ])
        with self.assertRaisesRegex(
                Exception,
                f"BR_Data_004 - The parameter df_measures contains '1' records duplicated, which is not expected. Make "
                f"sure your remove all duplicates for values"
        ):
            DatePredictabilityComputer()._check_and_cleanse_data(
                df_measures=the_df_client_measures,
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import unittest
from inspect import stack
from logging import getLogger, Logger

from com_enovation.toolbox.excel_dashboard.cell_notation_helper import CellNotationHelper


class TestFunction__IsPatternXxx(unittest.TestCase):

    _logger: Logger = getLogger(__name__)

    _list_the_incorrect_patterns: list[str] = [
        " D3",
        "32D"
    ]

    _list_the_cell_patterns: list[str] = [
        "--A--1",
        "@-A--1",
        "@$A--1",
        "-$A--1",
        "--A@-1",
        "--A@$1",
        "--A-$1",
        "@$A@$1"
    ]

    _list_the_range_patterns: list[str] = [
        "--A--1:@-A--1",
        "@-A--1:@$A--1",
        "@$A--1:-$A--1",
        "-$A--1:--A@-1",
        "--A@-1:--A@$1",
        "--A@$1:--A-$1",
        "--A-$1:@$A@$1",
        "@$A@$1:--A--1"
    ]

    _list_the_column_patterns: list[str] = [
        "--A:--A",
        "@-A:--A",
        "@$A:--A",
        "-$A:--A",
        "--A:@-A",
        "--A:@$A",
        "--A:-$A",
        "@$A:@$A"
    ]

    _list_the_row_patterns: list[str] = [
        "--1:--1",
        "@-1:--1",
        "@$1:--1",
        "-$1:--1",
        "--1:@-1",
        "--1:@$1",
        "--1:-$1",
        "@$1:@$1"
    ]

    def test_function__is_pattern_cell(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_str_pattern in self._list_the_cell_patterns:
            self.assertTrue(
                CellNotationHelper.is_pattern_cell(str_pattern=i_str_pattern),
                msg=f"Error when process pattern '{i_str_pattern}', expected to be OK."
            )

        for i_str_pattern in \
                self._list_the_range_patterns + \
                self._list_the_column_patterns + \
                self._list_the_row_patterns:
            self.assertFalse(
                CellNotationHelper.is_pattern_cell(str_pattern=i_str_pattern),
                msg=f"Error when process pattern '{i_str_pattern}', expected to be NOK."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__is_pattern_range(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_str_pattern in self._list_the_range_patterns:
            self.assertTrue(
                CellNotationHelper.is_pattern_range(str_pattern=i_str_pattern),
                msg=f"Error when process pattern '{i_str_pattern}', expected to be OK."
            )

        for i_str_pattern in \
                self._list_the_cell_patterns + \
                self._list_the_column_patterns + \
                self._list_the_row_patterns:
            self.assertFalse(
                CellNotationHelper.is_pattern_range(str_pattern=i_str_pattern),
                msg=f"Error when process pattern '{i_str_pattern}', expected to be NOK."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__is_pattern_column(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_str_pattern in self._list_the_column_patterns:
            self.assertTrue(
                CellNotationHelper.is_pattern_column(str_pattern=i_str_pattern),
                msg=f"Error when process pattern '{i_str_pattern}', expected to be OK."
            )

        for i_str_pattern in \
                self._list_the_cell_patterns + \
                self._list_the_range_patterns + \
                self._list_the_row_patterns:
            self.assertFalse(
                CellNotationHelper.is_pattern_column(str_pattern=i_str_pattern),
                msg=f"Error when process pattern '{i_str_pattern}', expected to be NOK."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__is_pattern_row(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_str_pattern in self._list_the_row_patterns:
            self.assertTrue(
                CellNotationHelper.is_pattern_row(str_pattern=i_str_pattern),
                msg=f"Error when process pattern '{i_str_pattern}', expected to be OK."
            )

        for i_str_pattern in \
                self._list_the_cell_patterns + \
                self._list_the_range_patterns + \
                self._list_the_column_patterns:
            self.assertFalse(
                CellNotationHelper.is_pattern_row(str_pattern=i_str_pattern),
                msg=f"Error when process pattern '{i_str_pattern}', expected to be NOK."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__is_pattern_correct(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_str_pattern in \
                self._list_the_cell_patterns + \
                self._list_the_range_patterns + \
                self._list_the_column_patterns + \
                self._list_the_row_patterns:
            self.assertTrue(
                CellNotationHelper.is_pattern_correct(str_pattern=i_str_pattern),
                msg=f"Error when process pattern '{i_str_pattern}', expected to be OK."
            )

        for i_str_pattern in self._list_the_incorrect_patterns:
            self.assertFalse(
                CellNotationHelper.is_pattern_correct(str_pattern=i_str_pattern),
                msg=f"Error when process pattern '{i_str_pattern}', expected to be NOK."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_exception(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_pattern in self._list_the_incorrect_patterns:

            with self.assertRaisesRegex(Exception, f"Pattern '{i_pattern}' is incorrect, and not a cell!"):
                CellNotationHelper.is_pattern_cell(str_pattern=i_pattern),

            with self.assertRaisesRegex(Exception, f"Pattern '{i_pattern}' is incorrect, and not a range!"):
                CellNotationHelper.is_pattern_range(str_pattern=i_pattern),

            with self.assertRaisesRegex(Exception, f"Pattern '{i_pattern}' is incorrect, and not a column!"):
                CellNotationHelper.is_pattern_column(str_pattern=i_pattern),

            with self.assertRaisesRegex(Exception, f"Pattern '{i_pattern}' is incorrect, and not a row!"):
                CellNotationHelper.is_pattern_row(str_pattern=i_pattern),

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import os
import unittest
from pathlib import Path

from click.testing import CliRunner, Result
from logging import Logger, getLogger
from inspect import stack
from com_enovation.enov import enov


class TestExcelDashboarder(unittest.TestCase):

    _logger: Logger = getLogger(__name__)

    def test_excel_dashboarder(self):
        """
        Note: When calling function 'CliRunner.invoke', we add a tag "# noinspection PyTypeChecker" not to raise a
        warning due to function "invoke" expecting function that we call to be a "BaseCommand", which does not seem to
        be the case...
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # First, we cleanse the output file from previous run
        if Path(os.path.join(os.path.dirname(__file__), '02_actual_results.xls')).exists():
            os.remove(os.path.join(os.path.dirname(__file__), '02_actual_results.xls'))
        self.assertFalse(
            Path(os.path.join(os.path.dirname(__file__), '02_actual_results.xls')).exists(),
            f"We still have a file '{os.path.join(os.path.dirname(__file__), '02_actual_results.xls')}', that will "
            f"generate issue while executing the test..."
        )

        the_runner = CliRunner()

        # noinspection PyTypeChecker
        the_result: Result = the_runner.invoke(
            cli=enov,
            args=[
                # To get all the logs...
                '--verbose',

                # We load the Excel file
                'df-load-xls',
                '-c', '["Id", "Client", "Label", "Label Bis"]',
                os.path.join(os.path.dirname(__file__), '01_input_file.xls'),
                'df_data',

                # We load the config
                'dict-load-json',
                os.path.join(os.path.dirname(__file__), '01_config.json'),
                'dict_config',

                # We compute date predictability
                'dashboard',
                '--alias-parameters', '{"df_data": "df_data"}',
                '--parameters', '{"tokens": []}',
                'dict_config',
                os.path.join(os.path.dirname(__file__), '02_actual_results.xls')
            ],
            catch_exceptions=True
        )
        self.assertEqual(
            0,
            the_result.exit_code,
            f"Unexpected exit_code, with stdout: {the_result.stdout}"
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(os.path.dirname(__file__), '02_actual_results.xls')
            ),
            f"File '02_actual_results.xls' was not persisted..., with stdout: {the_result.stdout}"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import datetime
from inspect import stack
from logging import Logger, getLogger
from pathlib import Path
from types import NoneType

import pywintypes
import win32com.client
from pandas import DataFrame


class ExcelTables:
    _dict_tables: dict[str, DataFrame]

    def __init__(
            self,
            dict_tables: dict[str, DataFrame],
    ):
        self._dict_tables = dict_tables

    @property
    def dict_tables(self) -> dict[str, DataFrame]:
        return self._dict_tables


class XlsDataConnectionToDf:
    """
    Class that interacts with Excel to perform actions such as:
    - Load an Excel file, refresh all data connections, persist and export data into a dataframe
    - TODO.
    """

    # Class Properties
    _lst_authorized_extensions: list[str] = [".xls", ".xlsm", ".xlsx"]
    _logger: Logger = getLogger(__name__)

    def refresh_and_return_xls_connections(
            self,
            p_path: Path,
            b_persist: bool = True,
    ) -> ExcelTables:
        """
        Function that executes the following steps:
        - Open an Excel Spreadsheet from its path
        - Refresh all data connections
        - Persist the file
        - Export all the Tables, or the ones from the paramater l_tables, and return them as DataFrames.

        Note: when crashing, a zombie process might survive:
        - ps | grep EXCEL --> return the process id
        - kill <process id>

        Idea for later:
        - Could add logic to only refresh one table
        - https://stackoverflow.com/questions/34296591/how-do-i-make-an-excel-refreshall-wait-to-close-until-finished
            for sheet in workbook.Sheets:
                print(sheet.name)
                    for table in sheet.QueryTables:
                        print("Found a query table on %s called %s" % (sheet.name, table.name))

                        # i.e.: disable background query, and therefore cause Excel to 'wait' until it's done refreshing
                        table.BackgroundQuery = False

                        # This both refreshes the table, AND if the Refresh() method returns True
                        # (i.e.: refreshes successfully), tells you so.
                        if table.Refresh() == True:
                            print("Query table %s refreshed!" % table.name)
                        else:
                            print("Query table %s FAILED to refresh." % table.name)

        :param p_path: path to the Excel file that contains the connections/ queries to refresh and export
        :param b_persist: persist the Excel file once all connections were refreshed. True by default.
        :return: an instance of ExcelTables
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We check the parameter p_path
        if p_path.is_file() is False:
            raise Exception(f"Path '{p_path}' does not lead to an existing file.")
        if p_path.suffix not in self._lst_authorized_extensions:
            raise Exception(f"Path '{p_path}' does not appear to have an extension among the ones authorized: "
                            f"{', '.join(self._lst_authorized_extensions)}")

        # Start an instance of Excel
        # - Using Dispatch: if an Excel instance was running, the Quit action will close all the worksheets
        # - Using DispatchEx: we have an isolated instance of Excel, that we can even hide...
        # noinspection is required as "client" is not recognize as a member of win32com
        try:
            # noinspection PyUnresolvedReferences
            o_xlapp = win32com.client.DispatchEx("Excel.Application")
            self._logger.info(f"Excel application is launched")
        except Exception as an_exception:
            raise Exception(f"Calling win32com.client.DispatchEx('Excel.Application') raised an exception")\
                from an_exception

        # Open the workbook in said instance of Excel
        try:
            o_wb = o_xlapp.workbooks.open(str(p_path))
            self._logger.info(f"Excel file '{p_path}' is opened.")
        except Exception as an_exception:
            try:
                o_xlapp.Quit()
            except Exception as sub_exception:
                self._logger.error(f"Tried to quit Excel, but faced an exception.", sub_exception)

            raise Exception(f"Calling o_xlapp.workbooks.open(p_path='{p_path}') raised an exception")\
                from an_exception

        # Optional, e.g. if you want to debug
        # o_xlapp.Visible = True

        # Refresh all data connections.
        try:
            o_wb.RefreshAll()
            self._logger.info(f"All data connections/ queries are being refreshed.")
        except Exception as an_exception:
            try:
                o_wb.Close(SaveChanges=False)
                o_xlapp.Quit()
            except Exception as sub_exception:
                self._logger.error(f"Tried to close and quit Excel, but faced an exception.", sub_exception)

            raise Exception(f"Calling o_wb.RefreshAll() raised an exception")\
                from an_exception

        # Wait for refresh to be done
        try:
            o_xlapp.CalculateUntilAsyncQueriesDone()
            self._logger.info(f"All data connections/ queries are now refreshed.")
        except Exception as an_exception:
            try:
                o_wb.Close(SaveChanges=False)
                o_xlapp.Quit()
            except Exception as sub_exception:
                self._logger.error(f"Tried to close and quit Excel, but faced an exception.", sub_exception)

            raise Exception(f"Calling o_xlapp.CalculateUntilAsyncQueriesDone() raised an exception")\
                from an_exception

        # We get all objects across all sheets into an instance of ExcelTables
        try:
            o_the_return: ExcelTables = self._workbook_to_bean(o_wb)
            self._logger.info(f"Workbook data retrieved into bean.")
        except Exception as an_exception:
            try:
                o_wb.Close(SaveChanges=False)
                o_xlapp.Quit()
            except Exception as sub_exception:
                self._logger.error(f"Tried to close and quit Excel, but faced an exception.", sub_exception)

            raise Exception(f"Calling _workbook_to_bean(o_wb) raised an exception")\
                from an_exception

        # We persist the file
        if b_persist:
            try:
                o_wb.Save()
                self._logger.info(f"Excel file '{p_path}' is saved.")
            except Exception as an_exception:
                try:
                    o_wb.Close(SaveChanges=False)
                    o_xlapp.Quit()
                except Exception as sub_exception:
                    self._logger.error(f"Tried to close and quit Excel, but faced an exception.", sub_exception)

                raise Exception(f"Calling o_wb.Save() raised an exception")\
                    from an_exception

        # We close the workbook
        try:
            o_wb.Close(SaveChanges=False)
        except Exception as an_exception:
            try:
                o_xlapp.Quit()
            except Exception as sub_exception:
                self._logger.error(f"Tried to close Excel, but faced an exception.", sub_exception)

            raise Exception(f"Calling o_wb.Close(SaveChanges=False) raised an exception")\
                from an_exception

        o_xlapp.Quit()
        del o_xlapp

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return o_the_return

    def _workbook_to_bean(self, o_wb) -> ExcelTables:
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")
        _dict_tables: dict[str, DataFrame] = {}

        # We process each and every worksheet
        for i_worksheet in o_wb.Worksheets:

            # We process each and every table
            for i_object in i_worksheet.ListObjects:

                _tpl_the_extracted_values: tuple = i_object.Range.Value

                _df_the_data: DataFrame = self._range_values_to_dataframe(
                    tpl_of_tpl_data_incl_headers=_tpl_the_extracted_values,
                    b_has_headers=True
                )

                _dict_tables[i_object.Name] = _df_the_data

        _o_the_return: ExcelTables = ExcelTables(_dict_tables)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return _o_the_return

    def _range_values_to_dataframe(
            self,
            tpl_of_tpl_data_incl_headers: tuple,
            b_has_headers: bool = True
    ) -> DataFrame:
        """
        To generate a DataFrame, we need:
        - To transform the tuple of tuples into lists
        - To transcode pywin.types into native types...
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We transform the data from tuples of tuples into list of list
        _lst_of_lst_data: list = [
            [
                self._transcode_pywintypes(i_cell)
                for i_cell in i_row
            ]
            for i_row in tpl_of_tpl_data_incl_headers
        ]

        if b_has_headers:
            _df_the_return: DataFrame = DataFrame(
                data=_lst_of_lst_data[1:],
                columns=_lst_of_lst_data[0]
            )
        else:
            _df_the_return: DataFrame = DataFrame(
                data=_lst_of_lst_data
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return _df_the_return

    def _transcode_pywintypes(
            self,
            cell_value
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        if isinstance(cell_value, pywintypes.TimeType):

            # It might happen that float values are considered as datetime...
            # As example, cell_value.timestamp() for '7' would return a negative number, that would crash
            if cell_value.timestamp() <= 0:
                self._logger.error(
                    f"One cell value is wrongly retrieved as a date, but function timestamp() returns a negative "
                    f"value. This could come from a numeric value in the source, refreshed as a date in the query. "
                    f"Check the query that refreshes data, and search for wrong cast into datetime...")
                _the_return: None = None

            else:
                # noinspection PyArgumentList
                _the_return: datetime.datetime = datetime.datetime.fromtimestamp(
                    timestamp=cell_value.timestamp(),
                    # tz=cell_value.tzinfo
                )

                # _the_return = cell_value.tzinfo.fromutc(_the_return)

        elif isinstance(cell_value, str):
            _the_return: str = cell_value

        elif isinstance(cell_value, NoneType):
            _the_return: NoneType = cell_value

        elif isinstance(cell_value, float):
            _the_return: float = cell_value

        else:
            raise Exception(f"Unexpected type {type(cell_value)}")

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return _the_return



from datetime import datetime
from inspect import stack
from logging import Logger, getLogger
from unittest import TestCase

from jsonschema import ValidationError

from com_enovation.toolbox.excel_dashboard.config_checker import ConfigChecker


class TestWorkbookSetProperties(TestCase):

    _logger: Logger = getLogger(__name__)

    _dict_a_full_format: dict = {
        "title": "a title to test",
        "subject": "a subject to test",
        "author": "an author to test",
        "manager": "a manager to test",
        "company": "a company to test",
        "category": "a category to test",
        "keywords": "a, keyword, to, test",
        "created": datetime(2018, 1, 1).strftime("yyyy-mm-dd"),
        "comments": "a comment to test",
        "status":  "a status to test",
        "hyperlink_base":  "a hyperlink_base to test"
    }

    def test_01_successful(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/workbook/set_properties/properties",
            lst_widgets_json_schemas=[]
        )

        _the_config_checker.validate(
            dict_to_validate=self._dict_a_full_format
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_02_unexpected_tag(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/workbook/set_properties/properties",
            lst_widgets_json_schemas=[]
        )

        _the_altered_dict: dict = dict(self._dict_a_full_format)
        _the_altered_dict["jsg"] = "bouh"

        with self.assertRaisesRegex(
                ValidationError,
                f"Additional properties are not allowed \\('jsg' was unexpected\\)"
        ):
            _the_config_checker.validate(
                dict_to_validate=_the_altered_dict
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_03_empty(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/workbook/set_properties/properties",
            lst_widgets_json_schemas=[]
        )

        _the_config_checker.validate(
            dict_to_validate={}
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from inspect import stack

from numpy import int64
from pandas import DataFrame, Series, Index
from logging import Logger, getLogger


class DataCheckFeatureFactory:
    """
    TODO: comment

    IDEAS:
    - A feature to check regular expression
    - A feature to check types
    """

    # Instance properties (decorated with @property)
    @property
    def dict_feature_getters(self) -> dict[str, callable]:
        return self._dict_feature_getters

    # Constants
    const_fn_data_extract_checker__count: str = "count"
    const_fn_data_extract_checker__column_label: str = "column"

    const__feature_getter_function_prefix: str = "get_data_check_feature__"

    # Class properties
    _logger: Logger = getLogger(__name__)

    def __init__(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We initialize the catalog of feature getter by introspecting the class and searching for all function
        # that starts with "get_data_check_feature__"
        self._dict_feature_getters: dict[str, callable] = {
            i_fn_label[len(self.const__feature_getter_function_prefix):]:
                getattr(DataCheckFeatureFactory, i_fn_label)
            for i_fn_label in dir(DataCheckFeatureFactory)
            if (
                    callable(getattr(DataCheckFeatureFactory, i_fn_label))
                    and i_fn_label.startswith(self.const__feature_getter_function_prefix)
            )
        }
        self._logger.debug(f"DataCheckFeatureFactory initialized with '{len(self.dict_feature_getters)}' "
                           f"feature getters: '{', '.join(self.dict_feature_getters)}'.")

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def get_check_feature(
            self,
            str_feature_label: str,
            **kwargs
    ) -> callable:
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        if str_feature_label not in self.dict_feature_getters:
            self._logger.debug(f"There is not getter for the feature labelled '{str_feature_label}'. Only the "
                               f"following are available: '{', '.join(self.dict_feature_getters)}")
            return None

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return self.dict_feature_getters[str_feature_label](self, **kwargs)

    def get_check_features_labels(self) -> list[str]:
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _lst_the_return: list[str] = list(self.dict_feature_getters.keys())

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return _lst_the_return

    def get_data_check_feature__check_duplicate(
            self,
            columns: list[str] | str
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        def fn__check_duplicate(
                df_data_extract: DataFrame,
        ) -> Series | None:
            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

            # We first verify that the columns to check are effectively present in the dataframe...
            if isinstance(columns, str):
                if columns not in df_data_extract.columns:
                    raise Exception(f"Column '{columns}' is missing from the data extract to check.")
            else:
                for i_col in columns:
                    if i_col not in df_data_extract.columns:
                        raise Exception(f"Column '{i_col}' is missing from the data extract to check.")

            # We select all the duplicated records
            the_duplicates: Series | DataFrame = df_data_extract[columns][
                df_data_extract.duplicated(subset=columns, keep=False)
            ]

            # If no duplicated records, we just return null
            if len(the_duplicates.index) == 0:
                s_the_return: None = None

            # If duplicated records are in a Series
            elif isinstance(the_duplicates, Series):
                # CHECK: we only check one column...
                if not isinstance(columns, str):
                    raise Exception(f"Unexpected case: duplicated returned as Series, while columns is of type "
                                    f"'{type(columns)}'.")

                # We count the number of occurrence by the parameter "columns"
                # The parameter "columns" is the index, and we label the number of occurrence "count"
                s_the_return: Series = the_duplicates.value_counts()

                # We label properly the index and the column
                s_the_return.name = self.const_fn_data_extract_checker__count
                s_the_return.index.name = columns

            # If duplicated records are in a DataFrame
            elif isinstance(the_duplicates, DataFrame):
                s_the_return: Series = the_duplicates.groupby(by=columns).size()

                # We label properly the index and the column
                s_the_return.name = self.const_fn_data_extract_checker__count
                # NOT NEEDED: s_the_return.index.name = columns

            # Else, exception...
            else:
                raise Exception(
                    f"The function df_data_extract[columns][df_data_extract.duplicated(subset=columns, keep=False)]"
                    f"returned an instance of type '{type(the_duplicates)}', while only Series or DataFrame was "
                    f"expected."
                )

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

            return s_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return fn__check_duplicate

    def get_data_check_feature__check_null(
            self,
            columns: list[str] | str
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        def fn__check_null(
                df_data_extract: DataFrame,
        ) -> Series | None:
            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

            # We first verify that the columns to check are effectively present in the dataframe...
            if isinstance(columns, str):
                if columns not in df_data_extract.columns:
                    raise Exception(f"Column '{columns}' is missing from the data extract to check.")

                lst_columns: list[str] = [columns]
            else:
                for i_col in columns:
                    if i_col not in df_data_extract.columns:
                        raise Exception(f"Column '{i_col}' is missing from the data extract to check.")

                lst_columns: list[str] = columns

            # we instantiate the Series we will return with the columns containing null values
            s_the_return: Series = Series(
                data=[],
                index=Index(
                    data=[],
                    name=self.const_fn_data_extract_checker__column_label
                ),
                name=DataCheckFeatureFactory.const_fn_data_extract_checker__count,
                dtype=int64
            )

            # We loop through each and every column to check
            for i_col in lst_columns:
                int_the_null_count: int = df_data_extract[i_col].isna().sum()
                if int_the_null_count > 0:
                    s_the_return[i_col] = int_the_null_count

            # If no column contains null, we just return None
            if len(s_the_return.index) == 0:
                s_the_return: None = None

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

            return s_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return fn__check_null

    def get_data_check_feature__check_isin(
            self,
            columns: list[str] | str,
            values: list
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        if isinstance(columns, str):
            _lst_the_columns: list[str] = [columns]
        else:
            _lst_the_columns: list[str] = columns

        _set_the_authorized_values: set = set(values)

        def fn__check_isin(
                df_data_extract: DataFrame,
        ) -> Series | None:
            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

            _s_the_return: Series | None = Series(dtype='object')

            # For each column to check
            for i_col in _lst_the_columns:

                # We verify that the column exists in the dataframe..
                if i_col not in df_data_extract.columns:
                    self._logger.error(f"Column '{i_col}' is missing from the data extract to check.")

                else:

                    # We get the unique values in the column
                    _set_the_unique_values: set = set(df_data_extract[i_col].unique())

                    _set_the_unauthorized_values: set = _set_the_unique_values - _set_the_authorized_values

                    if len(_set_the_unauthorized_values) > 0:
                        _s_the_return[i_col] = list(_set_the_unauthorized_values)

            if len(_s_the_return) == 0:
                _s_the_return = None

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

            return _s_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return fn__check_isin



import shutil
from inspect import stack
from logging import Logger, getLogger
from pathlib import Path
import win32com.client


class XlsRefreshDataConnections:
    """
    Class that interacts with Excel to refresh data connections/ queries.
    """

    # Class Properties
    _lst_authorized_extensions: list[str] = [".xls", ".xlsm", ".xlsx"]
    _logger: Logger = getLogger(__name__)

    def refresh_data_connections(
            self,
            p_file: Path,
            p_save_as: Path = None,
            b_visible: bool = False
    ):
        """
        Function that executes the following steps:
        - Open an Excel Spreadsheet from its path
        - Refresh all data connections
        - Persist the file, or save as another file
        - Let Excel open in case b_visible is set to True (otherwise, Excel instance is not visible)

        Note: when crashing, a zombie process might survive:
        - ps | grep EXCEL --> return the process id
        - kill <process id>
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We check the parameter p_file
        if p_file.is_file() is False:
            raise Exception(f"Path p_file:'{p_file}' does not lead to an existing file.")
        if p_file.suffix not in self._lst_authorized_extensions:
            raise Exception(f"Path p_file:'{p_file}' does not appear to have an extension among the ones authorized: "
                            f"{', '.join(self._lst_authorized_extensions)}")

        if p_save_as is None:
            _p_the_path: Path = p_file

        else:
            # We check the parameter p_save_as
            if p_save_as.is_file() is True:
                raise Exception(f"Path p_save_as:'{p_save_as}' does lead to an existing file.")
            if p_file.suffix not in self._lst_authorized_extensions:
                raise Exception(f"Path p_save_as:'{p_save_as}' does not appear to have an extension among the ones "
                                f"authorized: {', '.join(self._lst_authorized_extensions)}")

            # We need to copy the file that will be refreshed, as it will be persisted!
            shutil.copyfile(
                src=p_file,
                dst=p_save_as
            )
            _p_the_path: Path = p_save_as

        # Start an instance of Excel
        # - Using Dispatch: if an Excel instance was running, the Quit action will close all the worksheets
        # - Using DispatchEx: we have an isolated instance of Excel, that we can even hide...
        # noinspection is required as "client" is not recognize as a member of win32com
        try:
            # noinspection PyUnresolvedReferences
            o_xlapp = win32com.client.DispatchEx("Excel.Application")
            self._logger.info(f"Excel application is launched")
        except Exception as an_exception:
            raise Exception(f"Calling win32com.client.DispatchEx('Excel.Application') raised an exception")\
                from an_exception

        # Open the workbook in said instance of Excel
        try:
            o_wb = o_xlapp.workbooks.open(str(_p_the_path))
            self._logger.info(f"Excel file '{_p_the_path}' is opened.")
        except Exception as an_exception:
            try:
                o_xlapp.Quit()
            except Exception as sub_exception:
                self._logger.error(f"Tried to quit Excel, but faced an exception.", sub_exception)

            raise Exception(f"Calling o_xlapp.workbooks.open(_p_the_path='{_p_the_path}') raised an exception")\
                from an_exception

        # Refresh all data connections.
        try:
            o_wb.RefreshAll()
            self._logger.info(f"All data connections/ queries are being refreshed.")
        except Exception as an_exception:
            try:
                o_wb.Close(SaveChanges=False)
                o_xlapp.Quit()
            except Exception as sub_exception:
                self._logger.error(f"Tried to close and quit Excel, but faced an exception.", sub_exception)

            raise Exception(f"Calling o_wb.RefreshAll() raised an exception")\
                from an_exception

        # Wait for refresh to be done
        try:
            o_xlapp.CalculateUntilAsyncQueriesDone()
            self._logger.info(f"All data connections/ queries are now refreshed.")
        except Exception as an_exception:
            try:
                o_wb.Close(SaveChanges=False)
                o_xlapp.Quit()
            except Exception as sub_exception:
                self._logger.error(f"Tried to close and quit Excel, but faced an exception.", sub_exception)

            raise Exception(f"Calling o_xlapp.CalculateUntilAsyncQueriesDone() raised an exception")\
                from an_exception

        if b_visible:
            o_xlapp.Visible = True

        else:

            # We save
            try:
                o_wb.Save()
                self._logger.info(f"Excel file '{_p_the_path}' is saved.")
            except Exception as an_exception:
                try:
                    o_wb.Close(SaveChanges=False)
                    o_xlapp.Quit()
                except Exception as sub_exception:
                    self._logger.error(f"Tried to close and quit Excel, but faced an exception.", sub_exception)

                raise Exception(f"Calling o_wb.Save() raised an exception")\
                    from an_exception

            # We close the workbook
            try:
                o_wb.Close(SaveChanges=False)
            except Exception as an_exception:
                try:
                    o_xlapp.Quit()
                except Exception as sub_exception:
                    self._logger.error(f"Tried to close Excel, but faced an exception.", sub_exception)

                raise Exception(f"Calling o_wb.Close(SaveChanges=False) raised an exception")\
                    from an_exception

            o_xlapp.Quit()
            del o_xlapp

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



from inspect import stack
from logging import Logger, getLogger
from unittest import TestCase

from com_enovation.toolbox.excel_dashboard.cell_notation_helper import CellNotationHelper


class TestFunction_TranslateXxxToReference(TestCase):

    _logger: Logger = getLogger(__name__)

    _lst_test_data_exceptions: list[str] = [
        "abc",
        "D:C",
        "3:4",
    ]

    _dict_test_data__range: dict = {
        "A1:C5": "A1:C5",

        "@A1:C5": "B1:C5",
        "A@1:C5": "A3:C5",
        "A1:@C5": "A1:D5",
        "A1:C@5": "A1:C7",
        "@A@1:@C@5": "B3:D7",

        # With dollars
        "@A1:C$5": "B1:C$5",
        "A@1:$C5": "A3:$C5",
        "A$1:@C5": "A$1:D5",
        "$A1:C@5": "$A1:C7",
        "@$A@$1:@$C@$5": "$B$3:$D$7",
    }

    _dict_test_data__column: dict = {
        "B:C": "B:C",
        "B:@C": "B:D",
        "@B:@C": "C:D",

        # With dollars
        "$B:C": "$B:C",
        "B:@$C": "B:$D",
        "@$B:@$C": "$C:$D",
    }

    _dict_test_data__row: dict = {
        "3:5": "3:5",
        "3:@5": "3:7",
        "@3:@5": "5:7",

        # With dollars
        "$3:5": "$3:5",
        "3:@$5": "3:$7",
        "@$3:@$5": "$5:$7",
    }

    _dict_test_data__cell: dict = {
        "C5": "C5",
        "@C5": "D5",
        "C@5": "C7",

        # With dollars
        "$C5": "$C5",
        "@C$5": "D$5",
        "$C@$5": "$C$7",
    }

    _dict_test_data__column_ref: dict = {
        "C": "C",
        "@C": "D",

        # With dollars
        "$C": "$C",
        "@$C": "$D",
    }

    _dict_test_data__row_ref: dict = {
        "5": "5",
        "@5": "7",

        # With dollars
        "$5": "$5",
        "@$5": "$7",
    }

    def test_function__translate_address_to_reference(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for k, v in (
                self._dict_test_data__range
                | self._dict_test_data__cell
                | self._dict_test_data__column
                | self._dict_test_data__row
        ).items():
            _str_result: str = CellNotationHelper.translate_address_to_reference(
                str_address=k,
                str_reference_cell="B3"
            )
            self.assertEqual(
                first=v,
                second=_str_result,
                msg=f"Function CellNotationHelper.translate_address_to_reference(str_reference_cell='B3', "
                    f"str_address='{k}') returned '{_str_result}' while we expected '{v}'."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__translate_range_to_reference(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for k, v in self._dict_test_data__range.items():
            _str_result: str = CellNotationHelper.translate_range_to_reference(
                str_range=k,
                str_reference_cell="B3"
            )
            self.assertEqual(
                first=v,
                second=_str_result,
                msg=f"Function CellNotationHelper.translate_range_to_reference(str_reference_address='B3', "
                    f"str_relative_address='{k}') returned '{_str_result}' while we expected '{v}'."
            )

        # We check exceptions for unexpected formats...
        for i_data in (
                self._dict_test_data__cell
                | self. _dict_test_data__column_ref
                | self. _dict_test_data__row_ref
                | self._dict_test_data__column
                | self._dict_test_data__row
        ):

            with self.assertRaises(Exception):
                CellNotationHelper.translate_range_to_reference(
                    str_range=i_data,
                    str_reference_cell="B3"
                )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__translate_column_to_reference(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for k, v in self._dict_test_data__column.items():
            _str_result: str = CellNotationHelper.translate_column_to_reference(
                str_column=k,
                str_reference_cell="B3"
            )
            self.assertEqual(
                first=v,
                second=_str_result,
                msg=f"Function CellNotationHelper.translate_column_to_reference(str_reference_address='B3', "
                    f"str_column='{k}') returned '{_str_result}' while we expected '{v}'."
            )

        # We check exceptions for unexpected formats...
        for i_data in (
                self._dict_test_data__cell
                | self. _dict_test_data__column_ref
                | self. _dict_test_data__row_ref
                | self._dict_test_data__row
        ):

            with self.assertRaises(Exception):
                CellNotationHelper.translate_column_to_reference(
                    str_column=i_data,
                    str_reference_cell="B3"
                )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__translate_row_to_reference(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for k, v in self._dict_test_data__row.items():
            _str_result: str = CellNotationHelper.translate_row_to_reference(
                str_row=k,
                str_reference_cell="B3"
            )
            self.assertEqual(
                first=v,
                second=_str_result,
                msg=f"Function CellNotationHelper.translate_row_to_reference(str_reference_address='B3', "
                    f"str_row='{k}') returned '{_str_result}' while we expected '{v}'."
            )

        # We check exceptions for unexpected formats...
        for i_data in (
                self._dict_test_data__cell
                | self. _dict_test_data__column_ref
                | self. _dict_test_data__row_ref
                | self._dict_test_data__column
        ):

            with self.assertRaises(Exception):
                CellNotationHelper.translate_row_to_reference(
                    str_row=i_data,
                    str_reference_cell="B3"
                )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__translate_cell_to_reference(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for k, v in self._dict_test_data__cell.items():
            _str_result: str = CellNotationHelper.translate_cell_to_reference(
                str_cell=k,
                str_reference_cell="B3"
            )
            self.assertEqual(
                first=v,
                second=_str_result,
                msg=f"Function CellNotationHelper.translate_cell_to_reference(str_reference_address='B3', "
                    f"str_relative_address='{k}') returned '{_str_result}' while we expected '{v}'."
            )

        # We check exceptions for unexpected formats...
        for i_data in (
                self._dict_test_data__range
                | self. _dict_test_data__column_ref
                | self. _dict_test_data__row_ref
                | self._dict_test_data__column
                | self._dict_test_data__row
        ):

            with self.assertRaises(Exception):
                CellNotationHelper.translate_cell_to_reference(
                    str_cell=i_data,
                    str_reference_cell="B3"
                )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__translate_column_ref_to_reference(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for k, v in self._dict_test_data__column_ref.items():
            _str_result: str = CellNotationHelper.translate_column_ref_to_reference(
                str_column_ref=k,
                str_reference_cell="B3"
            )
            self.assertEqual(
                first=v,
                second=_str_result,
                msg=f"Function CellNotationHelper.translate_cell_to_reference(str_reference_address='B3', "
                    f"str_relative_address='{k}') returned '{_str_result}' while we expected '{v}'."
            )

        # We check exceptions for unexpected formats...
        for i_data in (
                self._dict_test_data__range
                | self. _dict_test_data__cell
                | self. _dict_test_data__row_ref
                | self._dict_test_data__column
                | self._dict_test_data__row
        ):

            with self.assertRaises(Exception):
                CellNotationHelper.translate_column_ref_to_reference(
                    str_column_ref=i_data,
                    str_reference_cell="B3"
                )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__translate_row_ref_to_reference(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for k, v in self._dict_test_data__row_ref.items():
            _str_result: str = CellNotationHelper.translate_row_ref_to_reference(
                str_row_ref=k,
                str_reference_cell="B3"
            )
            self.assertEqual(
                first=v,
                second=_str_result,
                msg=f"Function CellNotationHelper.translate_row_ref_to_reference(str_reference_address='B3', "
                    f"str_relative_address='{k}') returned '{_str_result}' while we expected '{v}'."
            )

        # We check exceptions for unexpected formats...
        for i_data in (
                self._dict_test_data__range
                | self. _dict_test_data__cell
                | self. _dict_test_data__column_ref
                | self._dict_test_data__column
                | self._dict_test_data__row
        ):

            with self.assertRaises(Exception):
                CellNotationHelper.translate_row_ref_to_reference(
                    str_row_ref=i_data,
                    str_reference_cell="B3"
                )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_exceptions(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # ############################################################################################################ #
        # 1. For all incorrect addresses                                                                               #
        # ############################################################################################################ #

        for i_data in self._lst_test_data_exceptions:

            with self.assertRaises(Exception):
                CellNotationHelper.translate_range_to_reference(
                    str_range=i_data,
                    str_reference_cell="B3"
                )

            with self.assertRaises(Exception):
                CellNotationHelper.translate_cell_to_reference(
                    str_cell=i_data,
                    str_reference_cell="B3"
                )

            with self.assertRaises(Exception):
                CellNotationHelper.translate_column_ref_to_reference(
                    str_column_ref=i_data,
                    str_reference_cell="B3"
                )

            with self.assertRaises(Exception):
                CellNotationHelper.translate_row_ref_to_reference(
                    str_row_ref=i_data,
                    str_reference_cell="B3"
                )

        # ############################################################################################################ #
        # 2. Incorrect reference cell                                                                                  #
        # ############################################################################################################ #

        with self.assertRaisesRegex(Exception, f"The address 'JSG' is not recognized as an address, and its pattern "
                                               f"could not be determined."):
            CellNotationHelper.translate_range_to_reference(
                str_range="J3",
                str_reference_cell="JSG"
            )

        with self.assertRaisesRegex(Exception, f"The address 'JSG' is not recognized as an address, and its pattern "
                                               f"could not be determined."):
            CellNotationHelper.translate_cell_to_reference(
                str_cell="J3",
                str_reference_cell="JSG"
            )

        with self.assertRaisesRegex(Exception, f"The address 'JSG' is not recognized as an address, and its pattern "
                                               f"could not be determined."):
            CellNotationHelper.translate_column_ref_to_reference(
                str_column_ref="J3",
                str_reference_cell="JSG"
            )

        with self.assertRaisesRegex(Exception, f"The address 'JSG' is not recognized as an address, and its pattern "
                                               f"could not be determined."):
            CellNotationHelper.translate_row_ref_to_reference(
                str_row_ref="J3",
                str_reference_cell="JSG"
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import os
import unittest
from pathlib import Path
from logging import Logger, getLogger
from inspect import stack
from pandas import DataFrame, read_excel
from com_enovation.toolbox.data_handler.data_modifier.data_modification_feature_factory \
    import DataModificationFeatureFactory


class TestDataModificationFeatureFactory_TypeColumns(unittest.TestCase):
    _logger: Logger = getLogger(__name__)

    def test_type_columns__all_column(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the data extract
        df_data_extract: DataFrame = read_excel(
            Path(os.path.join(os.path.dirname(__file__),
                              '01.data_extract.xlsx')),
            dtype="object"
        )

        # Run 1: we type each and every column
        df_the_modified_dataframe: DataFrame = DataModificationFeatureFactory(). \
            get_data_modification_feature__type_columns(
            dict_columns={
                "Jsg - ID": "str",
                "Jsg - Snapshot": "str",
                "Jsg - Label": "str",
                "Jsg - Type": "str",
                "Jsg - Solution": "str",
                "Jsg - Created On": "datetime",
                "Jsg - Updated On": "date",
                "Jsg - Status": "str",
                "Jsg - Stage": "str",
                "Jsg - Conf Lvl": "float",
                "Jsg - Close Date": "datetime",
                "Jsg - Manager": "str",
                "Jsg - Director": "str",
                "Jsg - Costs": "int",
                "Jsg - Revenues": "int",
                "Jsg - Null Values A": "str",
                "Jsg - Null Values B": "str",
            }
        )(df_data_extract)

        self.assertEqual(
            first=str(df_the_modified_dataframe.dtypes),
            second="""Jsg - ID                       object
Jsg - Snapshot                 object
Jsg - Label                    object
Jsg - Type                     object
Jsg - Solution                 object
Jsg - Created On       datetime64[ns]
Jsg - Updated On               object
Jsg - Status                   object
Jsg - Stage                    object
Jsg - Conf Lvl                float32
Jsg - Close Date       datetime64[ns]
Jsg - Manager                  object
Jsg - Director                 object
Jsg - Costs                     int16
Jsg - Revenues                float64
Jsg - Null Values A            object
Jsg - Null Values B            object
dtype: object""",
            msg="Error when checking typed dataframe"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import os
import unittest
from pathlib import Path
from logging import Logger, getLogger
from inspect import stack

from pandas import DataFrame, read_excel
from pandas.testing import assert_frame_equal

from com_enovation.toolbox.data_handler.data_modifier.data_modification_feature_factory \
    import DataModificationFeatureFactory


class TestDataModificationFeatureFactory_RepeatLast(unittest.TestCase):
    _logger: Logger = getLogger(__name__)

    def test_repeat_last__vanilla(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the data extract
        df_data_extract: DataFrame = read_excel(
            Path(os.path.join(os.path.dirname(__file__), '01.test_data.xlsx'))
        )

        # We type the data extract...
        df_data_extract = DataModificationFeatureFactory(). \
            get_data_modification_feature__type_columns(
            dict_columns={
                "Key": "str",
                "Date": "date",
                "Measure": "int"
            }
        )(df_data_extract)

        df_the_modified_dataframe: DataFrame = DataModificationFeatureFactory(). \
            get_data_modification_feature__repeat_last(
                s_column_to_repeat="Measure",
                s_column_repeat="Last Measure",
                lst_key_columns=["Key"],
                b_copy=True
            )(df_data_extract[["Key", "Date", "Measure"]])

        assert_frame_equal(
            left=df_data_extract,
            right=df_the_modified_dataframe,
            check_dtype=False
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from inspect import stack
from logging import Logger, getLogger
from unittest import TestCase

from jsonschema import ValidationError

from com_enovation.toolbox.excel_dashboard.config_checker import ConfigChecker


class TestWorkbookAddFormat(TestCase):

    _logger: Logger = getLogger(__name__)

    _dict_a_full_format: dict = {
        # Category Font. Font used in the cell. Excel can only display fonts that are installed on the system
        # that it is running on. Therefore it is best to use the fonts that come as standard such as
        # ‘Calibri’, ‘Times New Roman’ and ‘Courier New’. The default font for an unformatted cell in Excel
        # 2007+ is ‘Calibri’.
        "font_name": "Times New Roman",

        # Category Font. Size of the font used in the cell. Excel adjusts the height of a row to accommodate
        # the largest font size in the row.
        "font_size": 15,

        # Category Font. Color of the font used in the cell. Color can be a Html style #RRGGBB string or a
        # limited number of named colors. See website for further details:
        # https://xlsxwriter.readthedocs.io/format.html.
        "font_color": "red",

        # Category Font. To turn on bold for the format font.
        "bold": False,

        # Category Font. To turn on italic for the format font.
        "italic": True,

        # Category Font. To turn on underline for the format. The available underline styles are: 1 = Single
        # underline (the default), 2 = Double underline, 33 = Single accounting underline and 34 = Double
        # accounting underline.
        "underline": 2,

        # Category Font. To set the strikeout property of the font.
        "font_strikeout": True,

        # Category Font. To set the superscript/subscript property of the font. The available options are: 1
        # = Superscript and 2 = Subscript
        "font_script": 1,

        # Category Number. Set the number format for a cell. It controls whether a number is displayed as an
        # integer, a floating point number, a date, a currency value or some other user defined format. The
        # numerical format of a cell can be specified by using a format string or an index to one of Excel’s
        # built-in formats: 'd mmm yyyy' = Format string or '0x0F' = Format index. See website for indices to
        # Excel's built-in formats: https://xlsxwriter.readthedocs.io/format.html
        "num_format": "dd-mùm-yyyy",

        # Category Protection. To set the cell locked state. This property can be used to prevent
        # modification of a cell’s contents. Following Excel’s convention, cell locking is turned on by
        # default. However, it only has an effect if the worksheet has been protected using the worksheet
        # protect() method.
        "locked": True,

        # Category Protection. To hide formulas in a cell. This property is used to hide a formula while
        # still displaying its result. This is generally used to hide complex calculations from end users who
        # are only interested in the result. It only has an effect if the worksheet has been protected using
        # the worksheet protect() method.
        "hidden": False,

        # Category Alignment. To set the horizontal alignment for data in the cell: left, center, right,
        # fill, justify, center_across or distributed
        "align": "right",

        # Category Alignment. To set the vertical alignment for data in the cell: top, vcenter, bottom,
        # vjustify or vdistributed
        "valign": "top",

        # Category Alignment. To set the rotation of the text in a cell. Rotation angle is in the range -90
        # to 90 and 270. The angle 270 indicates text where the letters run from top to bottom.
        "rotation": 90,

        # Category Alignment. To turn text wrapping on for text in a cell.
        "text_wrap": True,

        # Category Alignment. To set the reading order for the text in a cell. This is useful when creating
        # Arabic, Hebrew or other near or far eastern worksheets. It can be used in conjunction with the
        # Worksheet right_to_left() method to also change the direction of the worksheet.
        "reading_order": True,

        # Category Alignment. To turn on the justify last text property. Only applies to Far Eastern versions
        # of Excel.
        "text_justlast": True,

        # Category Alignment. To center text across adjacent cells. Text can be aligned across two or more
        # adjacent cells. This is an alias for align='center_across'). Only the leftmost cell should contain
        # the text. The other cells in the range should be blank but should include the formatting.
        "center_across": True,

        # Category Alignment. To set the cell text indentation level, as an integer.
        "indent": 4,

        # Category Alignment. To turn on the text 'shrink to fit' for a cell.
        "shrink": True,

        # Category Pattern. To set the background pattern of a cell. The most common pattern is 1 which is a
        # solid fill of the background color. Pattern ranges from 0 to 18
        "pattern": 5,

        # Category Pattern. To set the color of the background pattern in a cell. If a pattern has not been
        # defined then a solid fill pattern is used as the default. The color can be a Html style #RRGGBB
        # string or a limited number of named colors.
        "bg_color": "black",

        # Category Pattern. To set the color of the foreground pattern in a cell. The color can be a Html
        # style #RRGGBB string or a limited number of named colors.
        "fg_color": "#FFFFFF",

        # Category Border. To set the cell border style. A cell border is comprised of a border on the
        # bottom, top, left and right. These can be set to the same value or individually using the relevant
        # properties. The following shows the border styles sorted by XlsxWriter index number: 0 = None,
        # 1 = Continuous weight 1, 2 = Continuous weight 2, 3 = Dash weight 1, 4 = Dot weight 1,
        # 5 = Continuous weight 3, 6 = Double weight 3, 7 = Continuous weight 0, 8 = Dash weight 2,
        # 9 = Dash Dot weight 1, 10 = Dash Dot weight 2, 11 = Dash Dot Dot weight, 12 = Dash Dot Dot weight
        # 2, 13 = SlantDash Dot weight 2
        "border": 0,

        # Category Border. To set the cell bottom border style.
        "bottom": 6,

        # Category Border. To set the cell top border style.
        "top": 5,

        # Category Border. To set the cell left border style.
        "left": 6,

        # "Category Border. To set the cell right border style."
        "right": 7,

        # Category Border. To set the color of the cell border. A cell border is comprised of a border on the
        # bottom, top, left and right. These can be set to the same value or individually using the relevant
        # properties. The color can be a Html style #RRGGBB string or a limited number of named colors.
        "border_color": "black",

        # Category Border. To set the color of the bottom cell border.
        "bottom_color": "black",

        # Category Border. To set the color of the top cell border.
        "top_color": "black",

        # Category Border. To set the color of the left cell border.
        "left_color": "green",

        # Category Border. To set the color of the right cell border.
        "right_color": "gray"
    }

    def test_01_successful(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/workbook/add_format/properties",
            lst_widgets_json_schemas=[]
        )

        _the_config_checker.validate(
            dict_to_validate=self._dict_a_full_format
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_02_unexpected_tag(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/workbook/add_format/properties",
            lst_widgets_json_schemas=[]
        )

        _the_altered_dict: dict = dict(self._dict_a_full_format)
        _the_altered_dict["jsg"] = "bouh"

        with self.assertRaisesRegex(
                ValidationError,
                f"Additional properties are not allowed \\('jsg' was unexpected\\)"
        ):
            _the_config_checker.validate(
                dict_to_validate=_the_altered_dict
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_03_empty(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/workbook/add_format/properties",
            lst_widgets_json_schemas=[]
        )

        with self.assertRaisesRegex(
                ValidationError,
                f"does not have enough properties"
        ):
            _the_config_checker.validate(
                dict_to_validate={}
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_04_incorrect_color(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/workbook/add_format/properties",
            lst_widgets_json_schemas=[]
        )

        _the_altered_dict: dict = dict(self._dict_a_full_format)
        _the_altered_dict["font_color"] = "bouh"

        with self.assertRaisesRegex(
                ValidationError,
                f"'bouh' is not valid under any of the given schemas"
        ):
            _the_config_checker.validate(
                dict_to_validate=_the_altered_dict
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import json
import os
from copy import deepcopy
from inspect import stack
from logging import Logger, getLogger

from com_enovation.toolbox.data_handler.data_handler import DataHandler


class DataBroadcaster:
    """
    A class to distribute:
    - For a list of recipients
      - Filter the data
      - Produce relevant dashboards
      - Persist into file system.

    Data to broadcast are provided into dataframe(s), at least one, and are to be prepared ahead of calling the
    broadcaster.

    Configuration is of the kind:
    {
        "dashboards configurations": {
            "a label for a dashboard": {
                "description": "a description for the excel dashboard, facultative",
                "path": "path to the excel dashboard configuration file"
            },
            (...)
        },
        "distributions configurations": {

            "a label for a distribution": {

                "description": "a description for the distribution, facultative",
                "dashboard configuration": "a label for a dashboard configuration",

                "data handlers": {

                    "a label for a data set to handle": {
                        # Note: this label should be among the function parameters when calling "distribute"

                        "description": "a description of the logic across the below steps",

                        "steps": [
                            {
                                "description": "a description of the step",
                                "feature": "a label of a data handler feature",
                                "param": {
                                    (...),
                                    "b_copy": true | false
                                }
                            }
                        ]
                    }
                }
            }
        }
    }
    """

    # Instance properties (decorated with @property)

    # Constants
    # -- To configure the dashboards configurations
    _const_str__config_tag__dashboards_cfgs__as_dict: str = "dashboards configurations"
    _const_str__config_tag__dashboard_cfg__path__as_path: str = "path"
    _const_str__config_tag__dashboard_cfg__description__as_str: str = "description"
    #     ... The below node is added during the initialization, and contains the dashboard configuration that is read
    #         from the files @path
    _const_str__config_tag__dashboard_cfg__configuration__as_dict: str = "configuration"

    # -- To configure the distributions aliases
    _const_str__config_tag__distributions_cfgs__as_dict: str = "distributions configurations"
    _const_str__config_tag__distribution_cfg__description__as_str: str = "description"
    _const_str__config_tag__distribution_cfg__dashboard_cfg__as_str: str = "description"
    _const_str__config_tag__distribution_cfg__data_sets_handlers_cfgs__as_dict: str = "data sets handlers"

    # Class properties
    _logger: Logger = getLogger(__name__)

    def __init__(
            self,
            dict_config: dict,
    ):
        """

        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        self._dict_dashboards_configurations: dict = self._load_dashboards_configurations(
            dict_dashboards_configurations=dict_config.get(self._const_str__config_tag__dashboards_cfgs__as_dict, None)
        )

        self._data_handlers: dict[str, DataHandler] = self._instantiate_data_handlers(
            dict_distributions_configurations=dict_config.get(self._const_str__config_tag__distributions_cfgs__as_dict)
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returned")

    def _load_dashboards_configurations(
            self,
            dict_dashboards_configurations: dict
    ) -> dict:
        """
        Process the node "dashboards configurations", that is a node at the root of the DataBroadcaster configuration
        file.

        This node is composed of the kind:
        {
            "dashboards configurations": {
                "a label for a dashboard": {
                    "description": "a description for the excel dashboard, facultative",
                    "path": "path to the excel dashboard configuration file"
                },
                (...)
            },
            "distributions configurations": {...}
        }

        Business rules:
        - BR-01: the node "dashboards configurations" is mandatory
        - BR-02: there should be at least one dashboard configuration
        - BR-03: each dashboard configuration should contain a node "path" (to a configuration file)
        - BR-04: each "path" should lead to an existing file
        - BR-05: each configuration file should be a json file (that can be loaded as a dictionary)
        - BR-06: each dashboard configuration could have a node "description" that must be of type string
        - BR-07: each dashboard configuration does not have other nodes but "path" and "description"

        Logic:
        - For each "dashboard configuration"
        - Load the "dashboard configuration file"
        - And return a deep-copied dictionary.
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _dict_the_return: dict = deepcopy(dict_dashboards_configurations)

        # BR-01: the node "dashboards configurations" is mandatory
        if _dict_the_return is None:
            raise Exception(f"BR-01: The node '{self._const_str__config_tag__dashboards_cfgs__as_dict}' is mandatory.")

        # BR-02: there should be at least one dashboard configuration
        if len(_dict_the_return) == 0:
            raise Exception(f"BR-02: there should be at least one dashboard configuration.")

        # We loop through each dashboard configuration
        for k_dashboard_config, v_dashboard_config in _dict_the_return.items():

            # We get the node "path"
            _str_the_path: str = v_dashboard_config.get(self._const_str__config_tag__dashboard_cfg__path__as_path, None)

            # BR-03: each dashboard configuration should contain a node "path" (to a configuration file)
            if _str_the_path is None:
                raise Exception(
                    f"BR-03: dashboard configuration for '{k_dashboard_config}' should contain a node "
                    f"'{self._const_str__config_tag__dashboard_cfg__path__as_path}' (to a configuration file)"
                )

            # BR-04: each "path" should lead to an existing file
            if os.path.isfile(path=_str_the_path) is False:
                raise Exception(
                    f"BR-04: path '{_str_the_path}' (for dashboard configuration '{k_dashboard_config}') should lead "
                    f"to an existing file"
                )

            try:
                with open(file=_str_the_path, mode='r') as json_file:
                    _dict_the_dashboard_configuration: dict = json.load(fp=json_file)
            except Exception as an_exception:
                raise Exception(
                    f"BR-05: configuration file '{_str_the_path}' for dashboard '{k_dashboard_config}' should be a "
                    f"json file (that can be loaded as a dictionary)."
                ) from an_exception

            # We get the node description
            _str_the_description: str | None = \
                v_dashboard_config.get(self._const_str__config_tag__dashboard_cfg__description__as_str, None)

            # BR-06: each dashboard configuration could have a node "description" that must be of type string
            if _str_the_description is not None:
                if isinstance(_str_the_description, str) is False:
                    raise Exception(
                        f"BR-06: each dashboard configuration could have a node "
                        f"'{self._const_str__config_tag__dashboard_cfg__description__as_str}' that must be of type "
                        f"string (but which is of  type '{type(_str_the_description)}')."
                    )

            # BR-07: each dashboard configuration does not have other nodes but "path" and "description"
            _lst_unexpected_nodes: list[str] = list(
                set(v_dashboard_config.keys())
                - {
                    self._const_str__config_tag__dashboard_cfg__path__as_path,
                    self._const_str__config_tag__dashboard_cfg__description__as_str
                }
            )
            if len(_lst_unexpected_nodes) > 0:
                raise Exception(
                    f"BR-07: dashboard configuration (for '{k_dashboard_config}') does have unexpected nodes: "
                    f"'{', '.join(_lst_unexpected_nodes)}'."
                )

            v_dashboard_config[self._const_str__config_tag__dashboard_cfg__configuration__as_dict] = \
                _dict_the_dashboard_configuration

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return _dict_the_return

    @staticmethod
    def _instantiate_data_handlers(
            dict_distributions_configurations: dict
    ) -> dict[str, DataHandler]:
        """
        Instantiate as many data handlers as data sets to broadcast. These data handlers will then be called to check/
        modify data for each distribution.

        From the Broadcaster configuration below:
        {
            "dashboards configurations": {...},
            "distributions configurations": {

                "a label for a distribution, aka k_distribution_lbl in the logic below": {

                    "description": "a description for the distribution, facultative",
                    (...),

                    "data sets handlers": {

                        "a label for a data set to handle, aka k_data_set in the logic below": {
                            # Note: this label should be among the function parameters when calling "distribute"

                            "description": "a description of the logic across the below steps",

                            "steps": [
                                {
                                    "description": "a description of the step",
                                    "feature": "a label of a data handler feature",
                                    "param": {
                                        (...),
                                        "b_copy": true | false
                                    }
                                }
                            ]
                        }
                    }
                }
            }
        }

        We produce a data handler for the data set "a label for a data set to handle", from the below configuration:
        {
            "a label for a distribution": {
                # That corresponds to "a label for the sequence that can be run in isolation"

                "description": "a description of the logic across the below steps",
                "steps": [
                    {
                        "description": "a description of the step",
                        "feature": "a label of a data handler feature",
                        "param": {
                            (...),
                            "b_copy": true | false
                        }
                    }
                ]
            }
        }

        """
        DataBroadcaster._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _dict_the_return: dict[str, DataHandler] = {}
        _dict_the_data_sets_handlers_cfgs: dict[str, dict] = {}

        # For each distribution
        for k_distribution_lbl, v_distribution_cfg in dict_distributions_configurations.items():

            # For each data set to handle
            for k_data_set, v_data_set_handler_cfg in v_distribution_cfg.get(
                    DataBroadcaster._const_str__config_tag__distribution_cfg__data_sets_handlers_cfgs__as_dict,
                    {}
            ).items():
                # We get the existing data handler
                _dict_existing_dh: dict = _dict_the_data_sets_handlers_cfgs.get(
                    k_data_set,
                    {}
                )

                _dict_existing_dh[k_distribution_lbl] = v_data_set_handler_cfg

                # We update the repository of data handlers
                _dict_the_data_sets_handlers_cfgs[k_data_set] = _dict_existing_dh

        for k_data_set, v_dh_cfg in _dict_the_data_sets_handlers_cfgs.items():
            _dict_the_return[k_data_set] = DataHandler(dict_config=v_dh_cfg)

        DataBroadcaster._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return _dict_the_return

    def distribute(
            self,
    ):
        """
        on peut commencer à distribuer...
        """
        pass



import os
import unittest
from pathlib import Path
from logging import Logger, getLogger
from inspect import stack
from pandas import DataFrame, read_excel, Series, Index
from pandas.testing import assert_series_equal

from com_enovation.toolbox.data_handler.data_checker.data_check_feature_factory import DataCheckFeatureFactory


class TestDataCheckFeatureFactory_CheckNull(unittest.TestCase):
    _logger: Logger = getLogger(__name__)

    def test_check_null__1_column(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the data extract
        df_data_extract: DataFrame = \
            read_excel(Path(os.path.join(os.path.dirname(__file__), '01.data_extract.xlsx')))

        # Run 1: null values found in one column
        s_the_return: Series = DataCheckFeatureFactory().get_data_check_feature__check_null(
            columns="Jsg - Null Values A"
        )(df_data_extract)

        assert_series_equal(
            left=s_the_return.sort_index(),
            right=Series(
                data=[5],
                index=Index(
                    data=["Jsg - Null Values A"],
                    name=DataCheckFeatureFactory.const_fn_data_extract_checker__column_label
                ),
                name=DataCheckFeatureFactory.const_fn_data_extract_checker__count,
            ).sort_index(),
            obj="Check 1: null values found in one column.",
            check_exact=True
        )

        # Run 2: no null values...
        s_the_return: Series = DataCheckFeatureFactory().get_data_check_feature__check_null(
            columns="Jsg - ID"
        )(df_data_extract)

        self.assertIsNone(
            obj=s_the_return,
            msg="Run 2, no null value in column ID"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_check_null__multiple_columns(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the data extract
        df_data_extract: DataFrame = \
            read_excel(Path(os.path.join(os.path.dirname(__file__), '01.data_extract.xlsx')))

        # Run 1: null values found in several columns
        s_the_return: Series = DataCheckFeatureFactory().get_data_check_feature__check_null(
            columns=["Jsg - Null Values A", "Jsg - Null Values B", "Jsg - ID"]
        )(df_data_extract)

        assert_series_equal(
            left=s_the_return.sort_index(),
            right=Series(
                data=[5, 1],
                index=Index(
                    data=["Jsg - Null Values A", "Jsg - Null Values B"],
                    name=DataCheckFeatureFactory.const_fn_data_extract_checker__column_label
                ),
                name=DataCheckFeatureFactory.const_fn_data_extract_checker__count
            ).sort_index(),
            obj="Check 1: null values found in 2 columns.",
            check_exact=True
        )

        # Run 2: no null values found across several columns
        s_the_return: Series = DataCheckFeatureFactory().get_data_check_feature__check_null(
            columns=["Jsg - ID", "Jsg - Snapshot"]
        )(df_data_extract)

        self.assertIsNone(
            obj=s_the_return,
            msg="Run 2, no null value in columns ID nor snapshot"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import os
from inspect import stack
from unittest import TestCase
from logging import Logger, getLogger

from pandas import DataFrame, read_excel

from com_enovation.helper.pandas_dataframe_typer import PandasDataframeTyper


class Test01DataframeTyper(TestCase):

    _logger: Logger = getLogger(__name__)

    def test_01_dataframe_typer(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _df_TC1_to_type: DataFrame = read_excel(
            io=os.path.join(os.path.dirname(__file__), 'TC_A_00.DataframeToType.xls'),
            dtype="object",
        )

        self.assertTrue(
            expr=(_df_TC1_to_type.dtypes == "object").all(),
            msg="We expect dataframe to type containing only 'object' columns."
        )

        _df_TC1_typed: DataFrame = PandasDataframeTyper.type(
            df_to_type=_df_TC1_to_type,
            dict_columns_to_type={
                "col1_str": PandasDataframeTyper.str__type__str,
                "col3_date": PandasDataframeTyper.str__type__date,
                "col4_float": PandasDataframeTyper.str__type__float,
                "col5_int": PandasDataframeTyper.str__type__int,
            }
        )

        self.assertListEqual(
            list1=_df_TC1_typed.dtypes.tolist(),
            list2=[
                "string",
                "object",
                "datetime64[ns]",
                "Float64",
                "Int64",
                "object"
            ],
            msg="We expect dataframe to be properly retyped."
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def test_02_unexpected_type(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        with self.assertRaisesRegex(
                Exception,
                f"We don't know how to handle transco into 'jsg_unexpected_type'."
        ):
            PandasDataframeTyper.type(
                df_to_type=DataFrame(columns=['col1', 'col2']),
                dict_columns_to_type={
                    "col1": PandasDataframeTyper.str__type__str,
                    "col2": "jsg_unexpected_type",
                }
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def test_03_unexpected_column(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # First call: we do not want to be strict --> no exception raised
        PandasDataframeTyper.type(
            df_to_type=DataFrame(columns=['col1', 'col2']),
            dict_columns_to_type={
                "col1": PandasDataframeTyper.str__type__str,
                "col3": PandasDataframeTyper.str__type__int,
            },
            b_strict=False
        )

        # Second call: we want to be strict --> we raise an exception
        with self.assertRaisesRegex(
                Exception,
                f"When requesting to type a dataframe, we discovered the following discrepancies:"
                f"\n\t- Columns in dataframe, but missing in  dictionary: 'col2'"
                f"\n\t- Columns in dictionary, but missing in dataframe: 'col3'."
        ):
            PandasDataframeTyper.type(
                df_to_type=DataFrame(columns=['col1', 'col2']),
                dict_columns_to_type={
                    "col1": PandasDataframeTyper.str__type__str,
                    "col3": PandasDataframeTyper.str__type__int,
                },
                b_strict=True
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
